# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from argparse import Namespace
import os

from colcon_core.package_descriptor import PackageDescriptor
from colcon_core.package_selection import _add_package_selection_arguments
from colcon_core.package_selection import _check_package_selection_parameters
from colcon_core.package_selection import add_arguments
from colcon_core.package_selection import get_package_selection_extensions
from colcon_core.package_selection import get_packages
from colcon_core.package_selection import PackageSelectionExtensionPoint
from colcon_core.package_selection import select_package_decorators
from mock import Mock
from mock import patch
import pytest

from .entry_point_context import EntryPointContext


class Extension1(PackageSelectionExtensionPoint):
    pass


class Extension2(PackageSelectionExtensionPoint):
    pass


def test_get_package_selection_extensions():
    with EntryPointContext(extension1=Extension1, extension2=Extension2):
        extensions = get_package_selection_extensions()
        assert ['extension1', 'extension2'] == list(extensions.keys())


def add_dummy_arguments(parser):
    parser.add_argument('arg')


def test__add_package_selection_arguments():
    parser = Mock()
    with EntryPointContext(extension1=Extension1, extension2=Extension2):
        extensions = get_package_selection_extensions()

        # invalid return value
        extensions['extension1'].add_arguments = Mock(return_value=True)
        with patch(
            'colcon_core.package_selection.add_package_discovery_arguments'
        ) as add_package_discovery_arguments:
            with patch('colcon_core.package_selection.logger.error') as error:
                add_arguments(parser)
        assert add_package_discovery_arguments.call_count == 1
        # the raised assertion is catched and results in an error message
        assert error.call_count == 1
        assert len(error.call_args[0]) == 1
        assert error.call_args[0][0].startswith(
            "Exception in package selection extension 'extension1': "
            'add_arguments() should return None\n')

        # raise exception
        extensions['extension1'].add_arguments = Mock(
            side_effect=RuntimeError('custom exception'))
        # check that arguments can be added
        extensions['extension2'].add_arguments = Mock(
            side_effect=add_dummy_arguments)
        with patch('colcon_core.package_selection.logger.error') as error:
            _add_package_selection_arguments(parser)
        assert extensions['extension2'].add_arguments.call_count == 1
        # the raised exception is catched and results in an error message
        assert error.call_count == 1
        assert len(error.call_args[0]) == 1
        assert error.call_args[0][0].startswith(
            "Exception in package selection extension 'extension1': "
            'custom exception\n')


def test_get_packages():
    args = Namespace()
    d1 = PackageDescriptor('/some/path')
    d1.name = 'one'
    d2 = PackageDescriptor('/other/path')
    d2.name = 'two'
    with patch(
        'colcon_core.package_selection.discover_packages',
        return_value=[d1, d2]
    ):
        decos = get_packages(args)
    assert len(decos) == 2
    assert decos[0].descriptor.name == 'one'
    assert decos[0].selected is True
    assert decos[1].descriptor.name == 'two'
    assert decos[1].selected is True

    d2.name = 'one'
    with patch(
        'colcon_core.package_selection.discover_packages',
        return_value=[d1, d2]
    ):
        with pytest.raises(RuntimeError) as e:
            get_packages(args)
        assert 'Duplicate package names not supported:' in str(e.value)
        assert '- one:' in str(e.value)
        assert '- {sep}some{sep}path'.format(sep=os.sep) in str(e.value)
        assert '- {sep}other{sep}path'.format(sep=os.sep) in str(e.value)


def test__check_package_selection_parameters():
    args = Mock()
    pkg_names = Mock()

    with EntryPointContext(extension1=Extension1, extension2=Extension2):
        extensions = get_package_selection_extensions()

        # nothing wrong with the arguments
        _check_package_selection_parameters(args, pkg_names)

        # raise exception
        extensions['extension1'].check_parameters = Mock(
            side_effect=RuntimeError('custom exception'))
        extensions['extension2'].check_parameters = Mock(return_value=None)
        with patch('colcon_core.package_selection.logger.error') as error:
            _check_package_selection_parameters(args, pkg_names)
        assert extensions['extension2'].check_parameters.call_count == 1
        # the raised exception is catched and results in an error message
        assert error.call_count == 1
        assert len(error.call_args[0]) == 1
        assert error.call_args[0][0].startswith(
            "Exception in package selection extension 'extension1': custom "
            'exception\n')

        # invalid return value
        extensions['extension1'].check_parameters = Mock(return_value=True)
        extensions['extension2'].check_parameters.reset_mock()
        with patch('colcon_core.package_selection.logger.error') as error:
            _check_package_selection_parameters(args, pkg_names)
        assert extensions['extension2'].check_parameters.call_count == 1
        # the raised assertion is catched and results in an error message
        assert error.call_count == 1
        assert len(error.call_args[0]) == 1
        assert error.call_args[0][0].startswith(
            "Exception in package selection extension 'extension1': "
            'check_parameters() should return None\n')

        # select some packages
        extensions['extension1'].check_parameters = Mock(
            side_effect=SystemExit(1))
        with pytest.raises(SystemExit):
            _check_package_selection_parameters(args, pkg_names)


def select_some_packages(*, args, decorators):
    for i, decorator in enumerate(decorators):
        decorator.selected = bool(i % 2)


def test_select_package_decorators():
    args = Mock()
    deco1 = Mock()
    deco1.selected = True
    deco2 = Mock()
    deco2.selected = True
    decos = [deco1, deco2]

    with EntryPointContext(extension1=Extension1, extension2=Extension2):
        extensions = get_package_selection_extensions()

        # raise exception
        extensions['extension2'].select_packages = Mock(return_value=None)
        with patch('colcon_core.package_selection.logger.error') as error:
            select_package_decorators(args, decos)
        # the raised exception is catched and results in an error message
        assert error.call_count == 1
        assert len(error.call_args[0]) == 1
        assert error.call_args[0][0].startswith(
            "Exception in package selection extension 'extension1': \n")

        # invalid return value
        extensions['extension1'].select_packages = Mock(return_value=True)
        with patch('colcon_core.package_selection.logger.error') as error:
            select_package_decorators(args, decos)
        # the raised assertion is catched and results in an error message
        assert error.call_count == 1
        assert len(error.call_args[0]) == 1
        assert error.call_args[0][0].startswith(
            "Exception in package selection extension 'extension1': "
            'select_packages() should return None\n')

        # select some packages
        extensions['extension1'].select_packages = Mock(
            side_effect=select_some_packages)
        select_package_decorators(args, decos)
        assert not deco1.selected
        assert deco2.selected



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from colcon_core.event.command import Command
from colcon_core.event_handler import EventHandlerExtensionPoint
from colcon_core.logging import colcon_logger
from colcon_core.plugin_system import satisfies_version

logger = colcon_logger.getChild(__name__)


class LogCommandEventHandler(EventHandlerExtensionPoint):
    """
    Log a 'debug' message for each command.

    In order to see the invoked commands the log level needs to be changed to
    show messages with the severity 'debug'.

    The extension handles events of the following types:
    - :py:class:`colcon_core.event.command.Command`
    """

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            EventHandlerExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')

    def __call__(self, event):  # noqa: D102
        data = event[0]

        if isinstance(data, Command):
            logger.debug(data.to_string())



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import os

from colcon_core.package_augmentation import augment_packages
from colcon_core.package_augmentation \
    import get_package_augmentation_extensions
from colcon_core.package_augmentation import PackageAugmentationExtensionPoint
from colcon_core.package_augmentation import update_descriptor
from colcon_core.package_augmentation import update_metadata
from colcon_core.package_descriptor import PackageDescriptor
from mock import Mock
from mock import patch

from .entry_point_context import EntryPointContext


class Extension1(PackageAugmentationExtensionPoint):
    PRIORITY = 80


class Extension2(PackageAugmentationExtensionPoint):
    pass


def test_get_package_augmentation_extensions():
    with EntryPointContext(extension1=Extension1, extension2=Extension2):
        extensions = get_package_augmentation_extensions()
        assert ['extension2', 'extension1'] == \
            list(extensions.keys())


def augment_package_metadata_with_data(desc, *, additional_argument_names):
    if str(desc.path) == '/some/path'.replace('/', os.sep):
        desc.metadata['key'] = 'value'


def augment_package_metadata_with_path(desc, *, additional_argument_names):
    desc.metadata['path'] = desc.path


def augment_package_with_hook(desc, *, additional_argument_names):
    desc.hooks += additional_argument_names


def test_augment_packages():
    desc1 = PackageDescriptor('/some/path')
    desc2 = PackageDescriptor('/other/path')
    descs = {desc1, desc2}
    with EntryPointContext(extension1=Extension1, extension2=Extension2):
        extensions = get_package_augmentation_extensions()
        extensions['extension1'].augment_package = Mock(
            side_effect=augment_package_metadata_with_data)
        extensions['extension2'].augment_package = Mock(
            side_effect=augment_package_metadata_with_path)
        augment_packages(descs)
    assert len(desc1.metadata) == 2
    assert set(desc1.metadata.keys()) == {'key', 'path'}
    assert desc1.path == desc1.metadata['path']

    assert len(desc2.metadata) == 1
    assert set(desc2.metadata.keys()) == {'path'}
    assert desc2.path == desc2.metadata['path']

    # raise exception
    desc1 = PackageDescriptor('/some/path')
    desc2 = PackageDescriptor('/other/path')
    descs = {desc1, desc2}
    with EntryPointContext(extension1=Extension1, extension2=Extension2):
        extensions = get_package_augmentation_extensions()
    extensions['extension1'].augment_package = Mock(
        side_effect=augment_package_with_hook)
    with patch('colcon_core.package_augmentation.logger.error') as error:
        augment_packages(
            descs, additional_argument_names=['arg1', 'arg2'],
            augmentation_extensions=extensions)
    assert desc1.hooks == ['arg1', 'arg2']
    assert desc2.hooks == ['arg1', 'arg2']
    # the raised exception is catched and results in an error message
    assert error.call_count == 1
    assert len(error.call_args[0]) == 1
    assert error.call_args[0][0].startswith(
        "Exception in package augmentation extension 'extension2': \n")

    # invalid return value
    desc1.hooks = []
    desc2.hooks = []
    extensions['extension2'].augment_packages = Mock(return_value=False)
    with patch('colcon_core.package_augmentation.logger.error') as error:
        augment_packages(
            descs, additional_argument_names=['arg1', 'arg2'],
            augmentation_extensions=extensions)
    assert desc1.hooks == ['arg1', 'arg2']
    assert desc2.hooks == ['arg1', 'arg2']
    # the raised assertion is catched and results in an error message
    assert error.call_count == 1
    assert len(error.call_args[0]) == 1
    assert error.call_args[0][0].startswith(
        "Exception in package augmentation extension 'extension2': ")


def test_update_descriptor():
    desc = PackageDescriptor('/some/path')
    assert len(desc.dependencies) == 0
    assert len(desc.hooks) == 0
    assert len(desc.metadata) == 0

    data = {
        'build-dependencies': {'b1', 'b2'},
        'test-dependencies': {'t1'},
    }
    update_descriptor(desc, data)
    assert len(desc.dependencies) == 2
    assert 'build' in desc.dependencies.keys()
    assert desc.dependencies['build'] == {'b1', 'b2'}
    assert 'test' in desc.dependencies.keys()
    assert desc.dependencies['test'] == {'t1'}

    data = {
        'dependencies': {'d1'},
        'hooks': ['hook1', 'hook2'],
        'key': 'value',
    }
    update_descriptor(desc, data, additional_argument_names=['*'])
    assert len(desc.dependencies) == 3
    assert 'build' in desc.dependencies.keys()
    assert desc.dependencies['build'] == {'d1', 'b1', 'b2'}
    assert 'run' in desc.dependencies.keys()
    assert desc.dependencies['run'] == {'d1'}
    assert 'test' in desc.dependencies.keys()
    assert desc.dependencies['test'] == {'d1', 't1'}

    assert len(desc.hooks) == 2
    assert desc.hooks == ['hook1', 'hook2']

    assert len(desc.metadata) == 1
    assert 'key' in desc.metadata
    assert desc.metadata['key'] == 'value'

    data = {
        'other': 'value',
        'some': 'value',
    }
    update_descriptor(
        desc, data, additional_argument_names=['some', 'unknown'])
    assert len(desc.metadata) == 2
    assert 'other' not in desc.metadata
    assert 'some' in desc.metadata
    assert desc.metadata['some'] == 'value'


def test_update_metadata():
    desc = PackageDescriptor('/some/path')
    desc.name = 'name'
    assert len(desc.metadata) == 0

    update_metadata(desc, 'd', {1: 'one', 2: 'two'})
    assert len(desc.metadata) == 1
    assert 'd' in desc.metadata.keys()
    assert desc.metadata['d'] == {1: 'one', 2: 'two'}

    update_metadata(desc, 'd', {2: 'TWO', 3: 'THREE'})
    assert len(desc.metadata) == 1
    assert 'd' in desc.metadata.keys()
    assert desc.metadata['d'] == {1: 'one', 2: 'TWO', 3: 'THREE'}

    update_metadata(desc, 'l', [1, 2])
    assert len(desc.metadata) == 2
    assert 'l' in desc.metadata.keys()
    assert desc.metadata['l'] == [1, 2]

    update_metadata(desc, 'l', [2, 3])
    assert len(desc.metadata) == 2
    assert 'l' in desc.metadata.keys()
    assert desc.metadata['l'] == [1, 2, 2, 3]

    update_metadata(desc, 's', {1, 2})
    assert len(desc.metadata) == 3
    assert 's' in desc.metadata.keys()
    assert desc.metadata['s'] == {1, 2}

    update_metadata(desc, 's', {2, 3})
    assert len(desc.metadata) == 3
    assert 's' in desc.metadata.keys()
    assert desc.metadata['s'] == {1, 2, 3}

    with patch('colcon_core.package_augmentation.logger.warning') as warn:
        update_metadata(desc, 's', 'different type')
    warn.assert_called_once_with(
        "update package 'name' metadata 's' from value '{1, 2, 3}' to "
        "'different type'")
    assert len(desc.metadata) == 3
    assert 's' in desc.metadata.keys()
    assert desc.metadata['s'] == 'different type'

    with patch('colcon_core.package_augmentation.logger.warning') as warn:
        update_metadata(desc, 's', 'same type')
    assert warn.call_count == 0
    assert len(desc.metadata) == 3
    assert 's' in desc.metadata.keys()
    assert desc.metadata['s'] == 'same type'



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from colcon_core.event.output import StderrLine
from colcon_core.event.output import StdoutLine
from colcon_core.event_handler.console_direct import ConsoleDirectEventHandler
from mock import patch


def test_console_direct():
    extension = ConsoleDirectEventHandler()

    with patch('sys.stdout') as stdout:
        event = StdoutLine(b'bytes line')
        extension((event, None))
        assert stdout.buffer.write.call_count == 1
        event = StdoutLine('string line')
        extension((event, None))
        assert stdout.write.call_count == 1

        stdout.buffer.write.reset_mock()
        stdout.write.reset_mock()
        extension(('unknown', None))
        assert stdout.buffer.write.call_count == 0
        assert stdout.write.call_count == 0

    with patch('sys.stderr') as stderr:
        event = StderrLine(b'bytes line')
        extension((event, None))
        assert stderr.buffer.write.call_count == 1
        event = StderrLine('string line')
        extension((event, None))
        assert stderr.write.call_count == 1

        stderr.buffer.write.reset_mock()
        stderr.write.reset_mock()
        extension(('unknown', None))
        assert stderr.buffer.write.call_count == 0
        assert stderr.write.call_count == 0



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from colcon_core.event.command import Command
from colcon_core.event.command import CommandEnded
from colcon_core.event_handler.log_command import LogCommandEventHandler
from mock import patch


def test_console_direct():
    extension = LogCommandEventHandler()

    with patch('colcon_core.event_handler.log_command.logger.debug') as debug:
        event = Command(['executable'], cwd='/some/path')
        extension((event, None))
        assert debug.call_count == 1

        debug.reset_mock()
        event = CommandEnded(['executable'], cwd='/some/path', returncode=1)
        extension((event, None))
        assert debug.call_count == 1

        debug.reset_mock()
        extension(('unknown', None))
        assert debug.call_count == 0



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import os
import sys

from colcon_core.subprocess import escape_shell_argument


class Command:
    """An event containing an invoked command."""

    __slots__ = ('cmd', 'cwd', 'env', 'shell')

    def __init__(self, cmd, *, cwd, env=None, shell=False):
        """
        Construct a Command.

        :param cmd: the sequence of program arguments
        :param cwd: the working directory
        :param env: a dictionary with environment variables
        :param shell: whether to use the shell as the program to execute
        """
        self.cmd = cmd
        self.cwd = cwd
        self.env = env
        self.shell = shell

    def to_string(self):
        """
        Get a string describing how to invoke the command.

        The message includes the working directory, modifications to
        environment variable and the command including the arguments.
        """
        string = "Invoking command in '{self.cwd}': ".format_map(locals())
        string += self._get_env_string()
        string += self._get_cmd_string()
        return string

    def _get_env_string(self):
        # determine differences in environment
        env = {}
        for var_name, new_value in (self.env or {}).items():
            # ignore some environment variables
            if sys.platform != 'win32':
                if var_name in ('PWD', ):
                    continue

            org_value = os.environ.get(var_name, None)
            # skip environment variables with the same value
            if new_value == org_value:
                continue

            # platform specific variable syntax
            if sys.platform != 'win32':
                var = '${' + var_name + '}'
            else:
                var = '%' + var_name + '%'

            if not org_value:
                # added environment variable
                value = new_value
            elif new_value.startswith(org_value):
                # appended environment variable
                value = var + new_value[len(org_value):]
            elif new_value.endswith(org_value):
                # prepended environment variable
                value = new_value[:-len(org_value)] + var
            else:
                # otherwise modified environment variable
                value = new_value
            env[var_name] = value

        # append variable assignments necessary for custom environment
        string = ''
        if env:
            for name in sorted(env.keys()):
                value = env[name]
                string += '{name}={value} '.format_map(locals())
        return string

    def _get_cmd_string(self):
        return ' '.join([
            escape_shell_argument(c) if self.shell else c
            for c in self.cmd])


class CommandEnded(Command):
    """An event containing a finished command."""

    __slots__ = ('returncode', )

    def __init__(self, cmd, *, cwd, returncode, env=None, shell=False):
        """
        Construct a CommandEnded.

        :param cmd: the sequence of program arguments
        :param cwd: the working directory
        :param returncode: the returncode of the command
        :param env: a dictionary with environment variables
        :param shell: whether to use the shell as the program to execute
        """
        super().__init__(cmd, cwd=cwd, env=env, shell=shell)
        self.returncode = returncode

    def to_string(self):
        """Get a string describing the invoked command and its return code."""
        string = "Invoked command in '{self.cwd}' returned " \
            "'{self.returncode}': ".format_map(locals())
        string += self._get_env_string()
        string += self._get_cmd_string()
        return string



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import os

from colcon_core.entry_point import EXTENSION_POINT_GROUP_NAME
from colcon_core.entry_point import get_all_entry_points
from colcon_core.entry_point import get_entry_points
from colcon_core.entry_point import load_entry_point
from colcon_core.entry_point import load_entry_points
from mock import Mock
from mock import patch
import pytest

from .environment_context import EnvironmentContext


class Group1:
    name = 'group1'


class Group2:
    name = 'group2'


class Dist():

    def __init__(self, group_name, group):
        self._group_name = group_name
        self._group = group

    def __lt__(self, other):
        return self._group_name < other._group_name

    def get_entry_map(self):
        return self._group


class EntryPoint:
    pass


def iter_entry_points(*, group):
    if group == EXTENSION_POINT_GROUP_NAME:
        return [Group1, Group2]
    assert group == Group1.name
    ep1 = EntryPoint()
    ep1.name = 'extA'
    ep2 = EntryPoint()
    ep2.name = 'extB'
    return [ep1, ep2]


def working_set():
    return [
        Dist('group1', {
            'group1': {ep.name: ep for ep in iter_entry_points(group='group1')}
        }),
        Dist('group2', {'group2': {'extC': EntryPoint()}}),
        Dist('other_group', {'other_group': {'extD': EntryPoint()}}),
    ]


def test_all_entry_points():
    with patch(
        'colcon_core.entry_point.iter_entry_points',
        side_effect=iter_entry_points
    ):
        with patch(
            'colcon_core.entry_point.WorkingSet',
            side_effect=working_set
        ):
            # successfully load a known entry point
            assert set(get_all_entry_points().keys()) == {'group1', 'group2'}
            assert set(get_all_entry_points()['group1'].keys()) == \
                {'extA', 'extB'}
            assert len(get_all_entry_points()['group1']['extA']) == 2
            assert isinstance(
                get_all_entry_points()['group1']['extA'][1], EntryPoint)
            assert get_all_entry_points()['group1']['extA'][1] .group_name == \
                'group1'
            assert get_all_entry_points()['group1']['extA'][1] .name == 'extA'


def test_entry_point_blocklist():
    # successful loading of entry point without a blocklist
    with patch(
        'colcon_core.entry_point.iter_entry_points',
        side_effect=iter_entry_points
    ):
        with patch(
            'colcon_core.entry_point.WorkingSet',
            side_effect=working_set
        ):
            entry_points = get_entry_points('group1')
    assert 'extA' in entry_points.keys()
    entry_point = entry_points['extA']
    assert entry_point.group_name == 'group1'
    assert entry_point.name == 'extA'

    entry_point.load = Mock()
    assert isinstance(entry_point, EntryPoint)
    load_entry_point(entry_point)
    assert entry_point.load.call_count == 1

    # successful loading of entry point not in blocklist
    entry_point.load.reset_mock()
    with EnvironmentContext(COLCON_EXTENSION_BLOCKLIST=os.pathsep.join([
        'group1.extB', 'group2.extC'])
    ):
        load_entry_point(entry_point)
    assert entry_point.load.call_count == 1

    # entry point in a blocked group can't be loaded
    entry_point.load.reset_mock()
    with EnvironmentContext(COLCON_EXTENSION_BLOCKLIST='group1'):
        with pytest.raises(RuntimeError) as e:
            load_entry_point(entry_point)
        assert 'The entry point group name is listed in the environment ' \
            'variable' in str(e.value)
    assert entry_point.load.call_count == 0

    # entry point listed in the blocklist can't be loaded
    with EnvironmentContext(COLCON_EXTENSION_BLOCKLIST=os.pathsep.join([
        'group1.extA', 'group1.extB'])
    ):
        with pytest.raises(RuntimeError) as e:
            load_entry_point(entry_point)
        assert 'The entry point name is listed in the environment variable' \
            in str(e.value)
    assert entry_point.load.call_count == 0


# mock entry points
class EntryPointRaisingException:
    group_name = 'group'
    name = 'exception'

    def load(self):
        raise Exception('entry point raising exception')


class EntryPointRaisingRuntimeError:
    group_name = 'group'
    name = 'runtime_error'

    def load(self):
        raise RuntimeError('entry point raising runtime error')


class EntryPointSuccess:
    group_name = 'group'
    name = 'success'

    def load(self):
        pass


@patch(
    'colcon_core.entry_point.get_entry_points',
    return_value={
        EntryPointRaisingException.name: EntryPointRaisingException(),
        EntryPointRaisingRuntimeError.name: EntryPointRaisingRuntimeError(),
        EntryPointSuccess.name: EntryPointSuccess(),
    })
def test_load_entry_points_with_exception(_):
    with patch('colcon_core.entry_point.logger.error') as error:
        extensions = load_entry_points('group')
    # the entry point raising an exception different than a runtime error
    # results in an error message
    assert error.call_count == 1
    assert len(error.call_args[0]) == 1
    assert "Exception loading extension 'group.exception'" \
        in error.call_args[0][0]
    assert 'entry point raising exception' in error.call_args[0][0]
    # neither of the entry points was loaded successfully
    assert extensions == {'success': None}



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from contextlib import suppress
with suppress(ImportError):
    # needed before importing distutils
    # to avoid warning introduced in setuptools 49.2.0
    import setuptools  # noqa: F401
from distutils.sysconfig import get_python_lib
from pathlib import Path

from colcon_core import shell
from colcon_core.environment import EnvironmentExtensionPoint
from colcon_core.environment import logger
from colcon_core.plugin_system import satisfies_version


class PythonPathEnvironment(EnvironmentExtensionPoint):
    """Extend the `PYTHONPATH` variable to find Python modules."""

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            EnvironmentExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')

    def create_environment_hooks(self, prefix_path, pkg_name):  # noqa: D102
        hooks = []

        python_path = Path(get_python_lib(prefix=str(prefix_path)))
        logger.log(1, "checking '%s'" % python_path)
        if python_path.exists():
            rel_python_path = python_path.relative_to(prefix_path)
            hooks += shell.create_environment_hook(
                'pythonpath', prefix_path, pkg_name,
                'PYTHONPATH', str(rel_python_path), mode='prepend')

        return hooks



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from distutils.sysconfig import get_python_lib
from pathlib import Path
from tempfile import TemporaryDirectory

from colcon_core.environment.pythonpath import PythonPathEnvironment
from mock import patch


def test_pythonpath():
    extension = PythonPathEnvironment()

    with TemporaryDirectory(prefix='test_colcon_') as prefix_path:
        prefix_path = Path(prefix_path)
        with patch(
            'colcon_core.shell.create_environment_hook',
            return_value=['/some/hook', '/other/hook']
        ):
            # Python path does not exist
            hooks = extension.create_environment_hooks(prefix_path, 'pkg_name')
            assert len(hooks) == 0

            # Python path exists
            python_path = Path(get_python_lib(prefix=str(prefix_path)))
            python_path.mkdir(parents=True)
            hooks = extension.create_environment_hooks(prefix_path, 'pkg_name')
            assert len(hooks) == 2



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import argparse

from colcon_core.event_handler import add_event_handler_arguments
from colcon_core.event_handler import apply_event_handler_arguments
from colcon_core.event_handler import EventHandlerExtensionPoint
from colcon_core.event_handler import format_duration
from colcon_core.event_handler import get_event_handler_extensions
from mock import Mock
import pytest

from .entry_point_context import EntryPointContext


class Extension1(EventHandlerExtensionPoint):
    """Some doc block."""

    def __init__(self):
        super().__init__()
        self.enabled = False


class Extension2(EventHandlerExtensionPoint):
    """Other doc block."""

    PRIORITY = 90


class Extension3(EventHandlerExtensionPoint):
    pass


def test_extension_interface():
    extension = Extension1()
    with pytest.raises(NotImplementedError):
        extension(None)


def test_get_shell_extensions():
    with EntryPointContext(
        extension1=Extension1, extension2=Extension2, extension3=Extension3
    ):
        extensions = get_event_handler_extensions(context=None)
    assert list(extensions.keys()) == [
        'extension1', 'extension3', 'extension2']


def test_add_event_handler_arguments():
    parser = argparse.ArgumentParser()
    with EntryPointContext(
        extension1=Extension1, extension2=Extension2, extension3=Extension3
    ):
        add_event_handler_arguments(parser)
    text = parser.format_help()
    assert 'extension1- extension2+ extension3+' in text
    assert '* extension1:' in text
    assert 'Some doc block' in text
    assert '* extension2:' in text
    assert 'Other doc block' in text


def test_apply_event_handler_arguments():
    with EntryPointContext(
        extension1=Extension1, extension2=Extension2, extension3=Extension3,
    ):
        extensions = get_event_handler_extensions(context=None)
    assert extensions['extension1'].enabled is False
    assert extensions['extension2'].enabled is True
    assert extensions['extension3'].enabled is True

    extensions['extension3'].enabled = None
    args = Mock()
    args.event_handlers = ['extension1+', 'extension2-']
    apply_event_handler_arguments(extensions, args)
    assert extensions['extension1'].enabled is True
    assert extensions['extension2'].enabled is False
    assert extensions['extension3'].enabled is None


def test_format_duration():
    # seconds below 10 with two decimal points
    assert format_duration(0) == '0.00s'
    assert format_duration(0.001) == '0.00s'
    assert format_duration(0.004999) == '0.00s'
    assert format_duration(0.005) == '0.01s'
    assert format_duration(9.99) == '9.99s'
    assert format_duration(9.994999) == '9.99s'
    assert format_duration(9.995) == '9.99s'  # floating point imprecision
    # seconds between 10 and 60 with one decimal points
    assert format_duration(9.995001) == '10.0s'
    assert format_duration(10) == '10.0s'
    assert format_duration(59.94) == '59.9s'
    # seconds above one minute with no decimal points
    assert format_duration(59.95) == '1min 0s'
    assert format_duration(83.45) == '1min 23s'
    assert format_duration(119.49) == '1min 59s'
    assert format_duration(119.5) == '2min 0s'
    assert format_duration(3599.4) == '59min 59s'
    # seconds above one hour with no decimal points
    assert format_duration(3599.5) == '1h 0min 0s'
    assert format_duration(5025.123) == '1h 23min 45s'
    assert format_duration(3599999) == '999h 59min 59s'
    # zero fixed decimal point
    assert format_duration(1.5, fixed_decimal_points=0) == '2s'
    assert format_duration(12.345, fixed_decimal_points=0) == '12s'
    # one fixed decimal points
    assert format_duration(1.5, fixed_decimal_points=1) == '1.5s'
    assert format_duration(12.345, fixed_decimal_points=1) == '12.3s'
    assert format_duration(34.5, fixed_decimal_points=1) == '34.5s'
    assert format_duration(3599.4, fixed_decimal_points=1) == '59min 59.4s'
    assert format_duration(4984.5, fixed_decimal_points=1) == '1h 23min 4.5s'
    # raise for negative parameter
    with pytest.raises(ValueError):
        format_duration(-1.0)



# Copyright 2020 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import os
import sys
import types

import pytest


@pytest.hookimpl(hookwrapper=True, tryfirst=True)
def pytest_terminal_summary(terminalreporter, exitstatus=None, config=None):
    """Redirect the summary warnings to stderr when run within colcon."""
    summary_warnings = terminalreporter.summary_warnings

    def redirect_to_stderr(self):
        nonlocal summary_warnings
        tw = self._tw
        import _pytest.config
        self._tw = _pytest.config.create_terminal_writer(
            self.config, sys.stderr)
        summary_warnings()
        self._tw = tw

    if 'COLCON' in os.environ:
        terminalreporter.summary_warnings = types.MethodType(
            redirect_to_stderr, terminalreporter)
    yield



# Copyright 2016-2020 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from glob import glob
import os
import sys

from colcon_core.argument_default import is_default_value
from colcon_core.argument_default import wrap_default_value
from colcon_core.package_discovery import logger
from colcon_core.package_discovery import PackageDiscoveryExtensionPoint
from colcon_core.package_identification import identify
from colcon_core.package_identification import IgnoreLocationException
from colcon_core.plugin_system import satisfies_version


class PathPackageDiscovery(PackageDiscoveryExtensionPoint):
    """Check specific paths for packages."""

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            PackageDiscoveryExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')

    def has_default(self):  # noqa: D102
        return True

    def add_arguments(
        self, *, parser, with_default, single_path=False
    ):  # noqa: D102
        parser.add_argument(
            '--paths',
            nargs='*' if not single_path else '?',
            metavar='PATH',
            default=wrap_default_value(['.']) if with_default else None,
            help='The paths to check for a package. Use shell wildcards '
                 '(e.g. `src/*`) to select all direct subdirectories' +
                 (' (default: .)' if with_default else ''))

    def has_parameters(self, *, args):  # noqa: D102
        return not is_default_value(args.paths) and \
            bool(args.paths)

    def discover(self, *, args, identification_extensions):  # noqa: D102
        if args.paths is None:
            return set()

        # on Windows manually check for wildcards and expand them
        if sys.platform == 'win32':
            _expand_wildcards(args.paths)

        logger.log(1, 'PathPackageDiscovery.discover(%s)', args.paths)

        visited_paths = set()
        descs = set()
        for path in args.paths:
            real_path = os.path.realpath(path)
            # avoid recrawling same paths
            if real_path in visited_paths:
                continue
            visited_paths.add(real_path)

            try:
                result = identify(identification_extensions, path)
            except IgnoreLocationException:
                continue
            if result:
                descs.add(result)

        return descs


def _expand_wildcards(paths):
    """
    Expand wildcards explicitly.

    This is only necessary on Windows.

    :param list paths: The paths to update in place
    """
    i = 0
    while i < len(paths):
        path = paths[i]
        if '*' not in path:
            i += 1
            continue
        expanded_paths = [
            p for p in sorted(glob(path))
            if os.path.isdir(p)]
        logger.log(
            5, "PathPackageDiscovery.discover() expanding '%s' to %s",
            path, expanded_paths)
        paths[i:i + 1] = expanded_paths
        i += len(expanded_paths)



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import argparse
from collections import OrderedDict

from colcon_core.argument_parser.destination_collector \
    import DestinationCollectorDecorator


def test_destination_collector_decorator():
    parser = argparse.ArgumentParser()
    decorator = DestinationCollectorDecorator(parser)
    decorator.add_argument('positional')
    assert decorator.get_destinations() == {}

    decorator.add_argument('--option', action='store_true')
    assert decorator.get_destinations() == OrderedDict([('option', 'option')])

    group = decorator.add_mutually_exclusive_group()
    group.add_argument('--other-option', action='store_true')
    assert decorator.get_destinations() == OrderedDict([
        ('option', 'option'), ('other-option', 'other_option')])

    subparser = decorator.add_subparsers(title='subs', dest='dest')
    parser = subparser.add_parser('parser')
    parser.add_argument('--another-option', action='store_true')
    assert decorator.get_destinations() == OrderedDict([
        ('option', 'option'), ('other-option', 'other_option'),
        ('another-option', 'another_option')])
    assert decorator.get_destinations(recursive=False) == OrderedDict([
        ('option', 'option'), ('other-option', 'other_option')])



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import copy

from colcon_core.dependency_descriptor import DependencyDescriptor


def test_constructor():
    d = DependencyDescriptor('foo')
    assert d == 'foo'
    assert str(d) == 'foo'
    assert d.name == 'foo'
    assert len(d.metadata) == 0

    d = DependencyDescriptor('foo', metadata={'bar': 'baz'})
    assert d == 'foo'
    assert str(d) == 'foo'
    assert d.name == 'foo'
    assert len(d.metadata) == 1
    assert 'bar' in d.metadata
    assert d.metadata['bar'] == 'baz'

    d2 = copy.deepcopy(d)
    assert d.name == d2.name
    assert d.metadata == d2.metadata
    d.metadata['bar'] = 'baz baz'
    assert d.metadata != d2.metadata



# Copyright 2019 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from pathlib import Path

from colcon_core.plugin_system import satisfies_version
from colcon_core.shell import logger
from colcon_core.shell import ShellExtensionPoint
from colcon_core.shell.template import expand_template


class DsvShell(ShellExtensionPoint):
    """Generate `.dsv` files describing the intended environment change."""

    # the priority needs to be higher than the default for primary shells
    PRIORITY = 200

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(ShellExtensionPoint.EXTENSION_POINT_VERSION, '^2.0')

    def create_prefix_script(self, prefix_path, merge_install):  # noqa: D102
        pass

    def create_package_script(
        self, prefix_path, pkg_name, hooks
    ):  # noqa: D102
        pkg_env_path = prefix_path / 'share' / pkg_name / 'package.dsv'
        logger.info("Creating package descriptor '%s'" % pkg_env_path)
        expand_template(
            Path(__file__).parent / 'template' / 'package.dsv.em',
            pkg_env_path,
            {
                'hooks': hooks,
            })

    def create_hook_set_value(
        self, env_hook_name, prefix_path, pkg_name, name, value,
    ):  # noqa: D102
        hook_path = prefix_path / 'share' / pkg_name / 'hook' / \
            ('%s.dsv' % env_hook_name)
        logger.info("Creating environment descriptor '%s'" % hook_path)
        expand_template(
            Path(__file__).parent / 'template' / 'hook_set_value.dsv.em',
            hook_path,
            {
                'name': name,
                'value': value,
            })
        return hook_path

    def create_hook_prepend_value(
        self, env_hook_name, prefix_path, pkg_name, name, subdirectory,
    ):  # noqa: D102
        hook_path = prefix_path / 'share' / pkg_name / 'hook' / \
            ('%s.dsv' % env_hook_name)
        logger.info("Creating environment descriptor '%s'" % hook_path)
        expand_template(
            Path(__file__).parent / 'template' / 'hook_prepend_value.dsv.em',
            hook_path,
            {
                'type_': 'prepend-non-duplicate',
                'name': name,
                'value': subdirectory,
            })
        return hook_path



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import argparse
import logging
from pathlib import Path
from tempfile import TemporaryDirectory

from colcon_core.verb import check_and_mark_build_tool
from colcon_core.verb import check_and_mark_install_layout
from colcon_core.verb import get_verb_extensions
from colcon_core.verb import update_object
from colcon_core.verb import VerbExtensionPoint
from mock import patch
import pytest

from .entry_point_context import EntryPointContext


def test_verb_interface():
    assert hasattr(VerbExtensionPoint, 'EXTENSION_POINT_VERSION')
    interface = VerbExtensionPoint()
    interface.add_arguments(parser=None)
    with pytest.raises(NotImplementedError):
        interface.main(context=None)


class Extension1(VerbExtensionPoint):
    pass


class Extension2(VerbExtensionPoint):
    pass


def test_get_verb_extensions():
    with EntryPointContext(extension1=Extension1, extension2=Extension2):
        extensions = get_verb_extensions()
    assert list(extensions.keys()) == ['extension1', 'extension2']


def test_check_and_mark_build_tool():
    with TemporaryDirectory(prefix='test_colcon_') as base_path:
        base_path = Path(base_path)

        # create marker if it doesn't exist
        check_and_mark_build_tool(str(base_path))
        marker_path = base_path / '.built_by'
        assert marker_path.exists()
        assert marker_path.read_text().rstrip() == 'colcon'

        # create path and marker if it doesn't exist
        path = base_path / 'no_base'
        check_and_mark_build_tool(str(path))
        assert path.exists()
        marker_path = path / '.built_by'
        assert marker_path.exists()
        assert marker_path.read_text().rstrip() == 'colcon'

        # existing marker with same content
        path = base_path / 'existing_marker'
        path.mkdir()
        marker_path = path / '.built_by'
        marker_path.write_text('colcon')
        check_and_mark_build_tool(str(path))
        assert marker_path.exists()
        assert marker_path.read_text().rstrip() == 'colcon'

        # existing marker with different content
        marker_path.write_text('other')
        with pytest.raises(RuntimeError):
            check_and_mark_build_tool(str(path))


def test_check_and_mark_install_layout():
    with TemporaryDirectory(prefix='test_colcon_') as base_path:
        base_path = Path(base_path)

        # create marker if it doesn't exist
        check_and_mark_install_layout(str(base_path), merge_install=False)
        marker_path = base_path / '.colcon_install_layout'
        assert marker_path.exists()
        assert marker_path.read_text().rstrip() == 'isolated'

        # create path and marker if it doesn't exist
        path = base_path / 'no_base'
        check_and_mark_install_layout(str(path), merge_install=True)
        assert path.exists()
        marker_path = path / '.colcon_install_layout'
        assert marker_path.exists()
        assert marker_path.read_text().rstrip() == 'merged'

        # existing marker with same content
        check_and_mark_install_layout(str(path), merge_install=True)
        assert marker_path.exists()
        assert marker_path.read_text().rstrip() == 'merged'

        # existing marker with different content
        with pytest.raises(RuntimeError):
            check_and_mark_install_layout(str(path), merge_install=False)

        # install base which is a file
        with pytest.raises(RuntimeError):
            check_and_mark_install_layout(str(marker_path), merge_install=True)


def test_update_object():
    obj = argparse.Namespace()
    key = 'key'
    msg_args = ['package_name', 'argument_type', 'value_source']
    update_object(obj, key, 'foo', *msg_args)
    assert hasattr(obj, key)
    assert getattr(obj, key) == 'foo'

    # debug message when overwriting a value with the same type
    with patch('colcon_core.verb.logger.log') as log:
        update_object(obj, key, 'bar', *msg_args)
    assert getattr(obj, key) == 'bar'
    assert log.call_count == 1
    assert len(log.call_args[0]) == 2
    assert log.call_args[0][0] == 5
    assert 'overwrite' in log.call_args[0][1]

    # warning message when overwriting a value with a different type
    with patch('colcon_core.verb.logger.log') as log:
        update_object(obj, key, 42, *msg_args)
    assert getattr(obj, key) == 42
    assert log.call_count == 1
    assert len(log.call_args[0]) == 2
    assert log.call_args[0][0] == logging.WARNING
    assert 'overwrite' in log.call_args[0][1]

    # an existing dictionary value is updated with the passed dictionary
    obj = argparse.Namespace()
    update_object(obj, key, {'foo': 'foo', 'bar': 'bar'}, *msg_args)
    with patch('colcon_core.verb.logger.log') as log:
        update_object(obj, key, {'bar': 'BAR', 'baz': 'BAZ'}, *msg_args)
    assert getattr(obj, key) == {'foo': 'foo', 'bar': 'BAR', 'baz': 'BAZ'}
    assert log.call_count == 1
    assert len(log.call_args[0]) == 2
    assert log.call_args[0][0] == 5
    assert 'update' in log.call_args[0][1]

    # an existing list value is extended with the passed list
    obj = argparse.Namespace()
    update_object(obj, key, ['foo', 'bar'], *msg_args)
    with patch('colcon_core.verb.logger.log') as log:
        update_object(obj, key, ['bar', 'baz'], *msg_args)
    assert getattr(obj, key) == ['foo', 'bar', 'bar', 'baz']
    assert log.call_count == 1
    assert len(log.call_args[0]) == 2
    assert log.call_args[0][0] == 5
    assert 'extend' in log.call_args[0][1]



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

"""
Event classes have no common base class.

They are being pushed to the queue of the
:py:class:`colcon_core.event_reactor.EventReactor` and handled by instances of
the :py:class:`colcon_core.event_handler.EventHandlerExtensionPoint` interface.
"""



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0


class StdoutLine:
    """
    An event containing a single line of text intended for `stdout`.

    The line has a trailing newline.
    """

    __slots__ = ('line', )

    def __init__(self, line):
        """
        Construct a StdoutLine.

        :param bytes|str line: The line of text
        """
        self.line = line


class StderrLine:
    """
    An event containing a single line of text intended for `stderr`.

    The line has a trailing newline.
    """

    __slots__ = ('line', )

    def __init__(self, line):
        """
        Construct a StderrLine.

        :param bytes|str line: The line of text
        """
        self.line = line



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import os
from pathlib import Path

from colcon_core.package_descriptor import PackageDescriptor
import pytest


def test_constructor():
    d = PackageDescriptor('/some/path')
    assert d.path == Path('/some/path')
    assert d.type is None
    assert d.name is None
    assert len(d.dependencies.keys()) == 0
    assert len(d.hooks) == 0
    assert len(d.metadata.keys()) == 0


def test_identifies_package():
    d = PackageDescriptor('/some/path')
    assert not d.identifies_package()
    d.type = 'type'
    assert not d.identifies_package()
    d.type = None
    d.name = 'name'
    assert not d.identifies_package()
    d.type = 'type'
    assert d.identifies_package()


def test_get_dependencies():
    d1 = PackageDescriptor('/some/path')
    d1.name = 'self'
    d1.dependencies['build'].add('build-depend')
    d1.dependencies['build'].add('depend')
    d1.dependencies['run'].add('run-depend')
    d1.dependencies['run'].add('depend')
    assert d1.get_dependencies() == {'build-depend', 'run-depend', 'depend'}

    d1.dependencies['test'].add('self')
    assert d1.get_dependencies(categories=('build', )) == \
        {'build-depend', 'depend'}

    with pytest.raises(AssertionError) as e:
        d1.get_dependencies()
    assert "'self'" in str(e.value)


def test_get_recursive_dependencies():
    d = PackageDescriptor('/some/path')
    d.name = 'A'
    d.dependencies['build'].add('B')
    d.dependencies['run'].add('c')
    d.dependencies['test'].add('d')

    d1 = PackageDescriptor('/other/path')
    d1.name = 'B'
    d1.dependencies['build'].add('e')
    d1.dependencies['run'].add('F')
    d1.dependencies['test'].add('G')

    d2 = PackageDescriptor('/another/path')
    d2.name = 'd'

    d3 = PackageDescriptor('/yet-another/path')
    d3.name = 'F'
    d3.dependencies['build'].add('h')
    d3.dependencies['test'].add('G')
    d3.dependencies['test'].add('I')

    d4 = PackageDescriptor('/more/path')
    d4.name = 'G'
    d4.dependencies['test'].add('I')

    d5 = PackageDescriptor('/yet-more/path')
    d5.name = 'I'
    # circular dependencies should be ignored
    d5.dependencies['run'].add('A')

    rec_deps = d.get_recursive_dependencies(
        {d, d1, d2, d3, d4, d5},
        direct_categories=('build', 'run'),
        recursive_categories=('run', 'test'))
    assert rec_deps == {
        # direct dependencies
        'B',
        # recursive dependencies
        'F', 'G', 'I',
    }


def test_magic_methods():
    d1 = PackageDescriptor('/some/path')
    d1.type = 'custom-type'
    d1.name = 'custom-name'
    d2 = PackageDescriptor('/some/path')
    d2.type = 'custom-type'
    d2.name = 'other-name'
    assert d1 != d2
    assert hash(d1) != hash(d2)

    d2.name = 'custom-name'
    assert d1 == d2
    assert hash(d1) == hash(d2)

    d1.dependencies['build'].add('build-depend')
    d2.hooks.append('hook')
    d2.metadata['key'] = 'value'
    assert d1 == d2
    assert hash(d1) == hash(d2)

    d2.type = 'other-type'
    assert d1 != d2
    assert hash(d1) != hash(d2)

    d2.type = 'custom-type'
    assert d1 == d2
    assert hash(d1) == hash(d2)

    d2.path = Path('/other/path')
    assert d1 != d2

    # comparing with other types always returns False
    assert d1 != []


def test_str():
    d = PackageDescriptor('/some/path')
    d.type = 'custom-type'
    d.name = 'custom-name'
    d.dependencies['build'].add('build-depend')
    d.dependencies['run'].add('run-depend')
    d.hooks += ('hook-a', 'hook-b')
    d.metadata['key'] = 'value'
    s = str(d)
    assert s.startswith('{')
    assert s.endswith('}')
    assert 'path: ' in s
    assert '/some/path'.replace('/', os.sep) in s
    assert 'type: ' in s
    assert 'custom-type' in s
    assert 'name: ' in s
    assert 'custom-name' in s
    assert 'dependencies: ' in s
    assert 'build-depend' in s
    assert 'run-depend' in s
    assert 'hooks: ' in s
    assert 'hook-a' in s
    assert 'metadata: ' in s
    assert 'value' in s



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from pathlib import Path
import sys
from tempfile import TemporaryDirectory

from colcon_core.shell.template import expand_template
from em import TransientParseError
from mock import patch
import pytest


def test_expand_template():
    with TemporaryDirectory(prefix='test_colcon_') as base_path:
        template_path = Path(base_path) / 'template.em'
        destination_path = Path(base_path) / 'expanded_template'

        # invalid template, missing @[end if]
        template_path.write_text(
            '@[if True]')
        with pytest.raises(TransientParseError):
            with patch('colcon_core.shell.template.logger.error') as error:
                expand_template(template_path, destination_path, {})
        # the raised exception is catched and results in an error message
        assert error.call_count == 1
        assert len(error.call_args[0]) == 1
        assert error.call_args[0][0].endswith(
            " processing template '{template_path}'".format_map(locals()))
        assert not destination_path.exists()

        # missing variable
        template_path.write_text(
            '@(var)')
        with pytest.raises(NameError):
            with patch('colcon_core.shell.template.logger.error') as error:
                expand_template(template_path, destination_path, {})
        # the raised exception is catched and results in an error message
        assert error.call_count == 1
        assert len(error.call_args[0]) == 1
        assert error.call_args[0][0].endswith(
            " processing template '{template_path}'".format_map(locals()))
        assert not destination_path.exists()

        # skip all symlink tests on Windows for now
        if sys.platform == 'win32':  # pragma: no cover
            return

        # remove destination if it is a symlink
        destination_path.symlink_to(template_path)
        assert destination_path.is_symlink()
        expand_template(template_path, destination_path, {'var': 'value'})
        assert not destination_path.is_symlink()
        assert destination_path.exists()



# Copyright 2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import copy


class DependencyDescriptor(str):
    """
    A descriptor for a package dependency.

    Beside the name of the dependency the 'metadata' dictionary can store any
    additional information.

    A dependency is identified by its name.
    The 'metadata' dictionary can store any additional information.

    Currently the class is inheriting from str for backwards compatibility.
    You should not rely on this but use the `name` property instead.
    """

    @staticmethod
    def __new__(cls, name, *, metadata=None):  # noqa: D102
        return str.__new__(cls, name)

    def __init__(self, name, *, metadata=None):  # noqa: D107
        self.metadata = metadata if metadata is not None else {}

    @property
    def name(self):
        """
        Name of the dependency.

        The property exists for future compatibility when the base class is
        removed.

        :rtype: str
        """
        return str(self)

    def __deepcopy__(self, memo=None):  # noqa: D105
        # surprisingly this is significantly faster than the default
        if not self.metadata:
            # explicitly skipping the deep copy of an empty dict is also faster
            return DependencyDescriptor(self.name)
        return DependencyDescriptor(
            self.name, metadata=copy.deepcopy(self.metadata, memo=memo))



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from collections import OrderedDict
import os
import os.path
from pathlib import Path
import traceback

from colcon_core.argument_parser.destination_collector \
    import DestinationCollectorDecorator
from colcon_core.event.job import JobUnselected
from colcon_core.event_handler import add_event_handler_arguments
from colcon_core.executor import add_executor_arguments
from colcon_core.executor import execute_jobs
from colcon_core.executor import Job
from colcon_core.executor import OnError
from colcon_core.package_identification.ignore import IGNORE_MARKER
from colcon_core.package_selection import add_arguments \
    as add_packages_arguments
from colcon_core.package_selection import get_packages
from colcon_core.plugin_system import satisfies_version
from colcon_core.shell import get_shell_extensions
from colcon_core.task import add_task_arguments
from colcon_core.task import get_task_extension
from colcon_core.task import TaskContext
from colcon_core.verb import check_and_mark_build_tool
from colcon_core.verb import check_and_mark_install_layout
from colcon_core.verb import logger
from colcon_core.verb import update_object
from colcon_core.verb import VerbExtensionPoint


class BuildPackageArguments:
    """Arguments to build a specific package."""

    def __init__(self, pkg, args, *, additional_destinations=None):
        """
        Construct a BuildPackageArguments.

        :param pkg: The package descriptor
        :param args: The parsed command line arguments
        :param list additional_destinations: The destinations of additional
          arguments
        """
        super().__init__()
        self.path = os.path.abspath(
            os.path.join(os.getcwd(), str(pkg.path)))
        self.build_base = os.path.abspath(os.path.join(
            os.getcwd(), args.build_base, pkg.name))
        self.install_base = os.path.abspath(os.path.join(
            os.getcwd(), args.install_base))
        self.merge_install = args.merge_install
        if not args.merge_install:
            self.install_base = os.path.join(
                self.install_base, pkg.name)
        self.symlink_install = args.symlink_install
        self.test_result_base = os.path.abspath(os.path.join(
            os.getcwd(), args.test_result_base, pkg.name)) \
            if args.test_result_base else None

        # set additional arguments
        for dest in (additional_destinations or []):
            # from the command line
            if hasattr(args, dest):
                update_object(
                    self, dest, getattr(args, dest),
                    pkg.name, 'build', 'command line')
            # from the package metadata
            if dest in pkg.metadata:
                update_object(
                    self, dest, pkg.metadata[dest],
                    pkg.name, 'build', 'package metadata')


class BuildVerb(VerbExtensionPoint):
    """Build a set of packages."""

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(VerbExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')

    def add_arguments(self, *, parser):  # noqa: D102
        parser.add_argument(
            '--build-base',
            default='build',
            help='The base path for all build directories (default: build)')
        parser.add_argument(
            '--install-base',
            default='install',
            help='The base path for all install prefixes (default: install)')
        parser.add_argument(
            '--merge-install',
            action='store_true',
            help='Merge all install prefixes into a single location')
        parser.add_argument(
            '--symlink-install',
            action='store_true',
            help='Use symlinks instead of copying files where possible')
        parser.add_argument(
            '--test-result-base',
            help='The base path for all test results (default: --build-base)')
        parser.add_argument(
            '--continue-on-error',
            action='store_true',
            help='Continue other packages when a package fails to build '
                 '(packages recursively depending on the failed package are '
                 'skipped)')
        add_executor_arguments(parser)
        add_event_handler_arguments(parser)

        add_packages_arguments(parser)

        decorated_parser = DestinationCollectorDecorator(parser)
        add_task_arguments(decorated_parser, 'colcon_core.task.build')
        self.task_argument_destinations = decorated_parser.get_destinations()

    def main(self, *, context):  # noqa: D102
        check_and_mark_build_tool(context.args.build_base)
        check_and_mark_install_layout(
            context.args.install_base,
            merge_install=context.args.merge_install)

        self._create_paths(context.args)

        decorators = get_packages(
            context.args,
            additional_argument_names=self.task_argument_destinations,
            recursive_categories=('run', ))

        install_base = os.path.abspath(os.path.join(
            os.getcwd(), context.args.install_base))
        jobs, unselected_packages = self._get_jobs(
            context.args, decorators, install_base)

        on_error = OnError.interrupt \
            if not context.args.continue_on_error else OnError.skip_downstream

        def post_unselected_packages(*, event_queue):
            nonlocal unselected_packages
            names = [pkg.name for pkg in unselected_packages]
            for name in sorted(names):
                event_queue.put(
                    (JobUnselected(name), None))

        rc = execute_jobs(
            context, jobs, on_error=on_error,
            pre_execution_callback=post_unselected_packages)

        self._create_prefix_scripts(install_base, context.args.merge_install)

        return rc

    def _create_paths(self, args):
        self._create_path(args.build_base)
        self._create_path(args.install_base)

    def _create_path(self, path):
        path = Path(os.path.abspath(path))
        if not path.exists():
            path.mkdir(parents=True, exist_ok=True)
        ignore_marker = path / IGNORE_MARKER
        if not os.path.lexists(str(ignore_marker)):
            with ignore_marker.open('w'):
                pass

    def _get_jobs(self, args, decorators, install_base):
        jobs = OrderedDict()
        unselected_packages = set()
        for decorator in decorators:
            pkg = decorator.descriptor

            if not decorator.selected:
                unselected_packages.add(pkg)
                continue

            extension = get_task_extension('colcon_core.task.build', pkg.type)
            if not extension:
                logger.warning(
                    "No task extension to 'build' a '{pkg.type}' package"
                    .format_map(locals()))
                continue

            recursive_dependencies = OrderedDict()
            for dep_name in decorator.recursive_dependencies:
                dep_path = install_base
                if not args.merge_install:
                    dep_path = os.path.join(dep_path, dep_name)
                recursive_dependencies[dep_name] = dep_path

            package_args = BuildPackageArguments(
                pkg, args, additional_destinations=self
                .task_argument_destinations.values())
            ordered_package_args = ', '.join([
                ('%s: %s' % (repr(k), repr(package_args.__dict__[k])))
                for k in sorted(package_args.__dict__.keys())
            ])
            logger.debug(
                "Building package '{pkg.name}' with the following arguments: "
                '{{{ordered_package_args}}}'.format_map(locals()))
            task_context = TaskContext(
                pkg=pkg, args=package_args,
                dependencies=recursive_dependencies)

            job = Job(
                identifier=pkg.name,
                dependencies=set(recursive_dependencies.keys()),
                task=extension, task_context=task_context)

            jobs[pkg.name] = job
        return jobs, unselected_packages

    def _create_prefix_scripts(self, install_base, merge_install):
        extensions = get_shell_extensions()
        for priority in extensions.keys():
            extensions_same_prio = extensions[priority]
            for extension in extensions_same_prio.values():
                try:
                    retval = extension.create_prefix_script(
                        Path(install_base), merge_install)
                    assert retval is None, \
                        'create_prefix_script() should return None'
                except Exception as e:  # noqa: F841
                    # catch exceptions raised in shell extension
                    exc = traceback.format_exc()
                    logger.error(
                        'Exception in shell extension '
                        "'{extension.SHELL_NAME}': {e}\n{exc}"
                        .format_map(locals()))
                    # skip failing extension, continue with next one



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import asyncio
import logging
import signal
import sys
import traceback

from colcon_core.executor import ExecutorExtensionPoint
from colcon_core.executor import OnError
from colcon_core.logging import colcon_logger
from colcon_core.plugin_system import satisfies_version
from colcon_core.subprocess import new_event_loop
from colcon_core.subprocess import SIGINT_RESULT

logger = colcon_logger.getChild(__name__)


class SequentialExecutor(ExecutorExtensionPoint):
    """
    Process one package at a time.

    The sequence follows the topological ordering.
    """

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            ExecutorExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')

    def execute(self, args, jobs, *, on_error=OnError.interrupt):  # noqa: D102
        # avoid debug message from asyncio when colcon uses debug log level
        asyncio_logger = logging.getLogger('asyncio')
        asyncio_logger.setLevel(logging.INFO)

        rc = 0
        loop = new_event_loop()
        asyncio.set_event_loop(loop)
        jobs = jobs.copy()
        try:
            while jobs:
                name, job = jobs.popitem(last=False)
                coro = job()
                future = asyncio.ensure_future(coro, loop=loop)
                try:
                    logger.debug(
                        "run_until_complete '{name}'".format_map(locals()))
                    loop.run_until_complete(future)
                except KeyboardInterrupt:
                    logger.debug(
                        "run_until_complete '{name}' was interrupted"
                        .format_map(locals()))
                    # override job rc with special SIGINT value
                    job.returncode = SIGINT_RESULT
                    # ignore further SIGINTs
                    signal.signal(signal.SIGINT, signal.SIG_IGN)
                    # wait for job which has also received a SIGINT
                    if not future.done():
                        logger.debug(
                            "run_until_complete '{name}' again"
                            .format_map(locals()))
                        loop.run_until_complete(future)
                        assert future.done()
                    # read potential exception to avoid asyncio error
                    _ = future.exception()  # noqa: F841
                    logger.debug(
                        "run_until_complete '{name}' finished"
                        .format_map(locals()))
                    return signal.SIGINT
                except Exception as e:  # noqa: F841
                    exc = traceback.format_exc()
                    logger.error(
                        "Exception in job execution '{name}': {e}\n{exc}"
                        .format_map(locals()))
                    return 1
                result = future.result()
                logger.debug(
                    "run_until_complete '{name}' finished with '{result}'"
                    .format_map(locals()))
                if result:
                    if not rc:
                        rc = result
                    if on_error in (OnError.interrupt, OnError.skip_pending):
                        # skip pending jobs
                        return rc
                    if on_error == OnError.skip_downstream:
                        # skip downstream jobs of failed one
                        for pending_name, pending_job in list(jobs.items()):
                            if job.identifier in pending_job.dependencies:
                                del jobs[pending_name]

        finally:
            try:
                # new in Python 3.7
                all_tasks = asyncio.all_tasks
            except AttributeError:
                all_tasks = asyncio.Task.all_tasks
            for task in all_tasks(loop):
                if not task.done():
                    logger.error("Task '{task}' not done".format_map(locals()))
            # HACK on Windows closing the event loop seems to hang after Ctrl-C
            # even though no futures are pending
            if sys.platform != 'win32':
                logger.debug('closing loop')
                loop.close()
                logger.debug('loop closed')
        return rc



# Copyright 2020 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from colcon_core.argument_default import is_default_value
from colcon_core.argument_default import unwrap_default_value
from colcon_core.argument_default import wrap_default_value
import pytest


def test_argument_default():
    values = [
        True,
        [1, 2, 3],
        'foo',
    ]
    for value in values:
        assert not is_default_value(value)
        with pytest.raises(ValueError):
            unwrap_default_value(value)
        default_value = wrap_default_value(value)
        assert is_default_value(default_value)
        assert type(default_value) != type(value)
        with pytest.raises(ValueError):
            wrap_default_value(default_value)
        unwrapped_value = unwrap_default_value(default_value)
        assert value == unwrapped_value

    value = 42
    unchanged_value = wrap_default_value(value)
    assert type(unchanged_value) == type(value)
    assert unchanged_value == value



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from collections import OrderedDict
import os
from pathlib import Path
from tempfile import TemporaryDirectory

from colcon_core.plugin_system import SkipExtensionException
from colcon_core.shell import check_dependency_availability
from colcon_core.shell import create_environment_hook
from colcon_core.shell import find_installed_packages
from colcon_core.shell import find_installed_packages_in_environment
from colcon_core.shell import get_colcon_prefix_path
from colcon_core.shell import get_command_environment
from colcon_core.shell import get_environment_variables
from colcon_core.shell import get_shell_extensions
from colcon_core.shell import ShellExtensionPoint
from mock import Mock
from mock import patch
import pytest

from .entry_point_context import EntryPointContext
from .environment_context import EnvironmentContext
from .run_until_complete import run_until_complete


class Extension1(ShellExtensionPoint):
    PRIORITY = 90


class Extension2(ShellExtensionPoint):
    pass


def test_extension_interface():
    extension = Extension1()
    with pytest.raises(NotImplementedError):
        extension.create_prefix_script(None, None)
    with pytest.raises(NotImplementedError):
        extension.create_package_script(None, None, None)
    with pytest.raises(NotImplementedError):
        extension.create_hook_set_value(None, None, None, None, None)
    with pytest.raises(NotImplementedError):
        extension.create_hook_append_value(None, None, None, None, None)
    with pytest.raises(NotImplementedError):
        extension.create_hook_prepend_value(None, None, None, None, None)
    with pytest.raises(NotImplementedError):
        extension.create_hook_include_file(None, None, None, None)

    coroutine = extension.generate_command_environment(None, None, None)
    with pytest.raises(NotImplementedError):
        run_until_complete(coroutine)


def test_get_shell_extensions():
    with EntryPointContext(extension1=Extension1, extension2=Extension2):
        extensions = get_shell_extensions()
    assert list(extensions.keys()) == [100, 90]
    assert list(extensions[100].keys()) == ['extension2']
    assert list(extensions[90].keys()) == ['extension1']


async def generate_command_environment(task_name, build_base, dependencies):
    return {'key': 'value'}


def test_get_command_environment():
    with EntryPointContext(extension1=Extension1, extension2=Extension2):
        extensions = get_shell_extensions()

        # one not implemented, one skipped extension
        extensions[90]['extension1'].generate_command_environment = Mock(
            side_effect=SkipExtensionException())
        coroutine = get_command_environment(None, '/build/base', None)
        with patch('colcon_core.shell.logger.debug') as debug:
            with patch('colcon_core.shell.logger.info') as info:
                with pytest.raises(RuntimeError) as e:
                    run_until_complete(coroutine)
        assert 'Could not find a shell extension for the command environment' \
            in str(e.value)
        assert extensions[90]['extension1'].generate_command_environment \
            .call_count == 1
        # the raised exceptions are catched and result in a debug/info message
        assert debug.call_count == 1
        assert len(debug.call_args[0]) == 1
        assert debug.call_args[0][0] == \
            "Skip shell extension 'extension2' for command environment"
        assert info.call_count == 1
        assert len(info.call_args[0]) == 1
        assert info.call_args[0][0].startswith(
            "Skip shell extension 'extension1' for command environment: ")

        # raise runtime error
        extensions[100]['extension2'].generate_command_environment = Mock(
            side_effect=RuntimeError('custom exception'))
        extensions[90]['extension1'].generate_command_environment.reset_mock()
        coroutine = get_command_environment(None, '/build/base', None)
        with pytest.raises(RuntimeError) as e:
            run_until_complete(coroutine)
        assert str(e.value) == 'custom exception'
        assert extensions[90]['extension1'].generate_command_environment \
            .call_count == 0

        # one exception, one successful
        extensions[100]['extension2'].generate_command_environment = Mock(
            side_effect=Exception('custom exception'))
        extensions[90]['extension1'].generate_command_environment = Mock(
            side_effect=generate_command_environment)
        coroutine = get_command_environment(None, '/build/base', None)
        with patch('colcon_core.shell.logger.error') as error:
            env = run_until_complete(coroutine)
        assert env == {'key': 'value'}
        # the raised exception is catched and results in an error message
        assert error.call_count == 1
        assert len(error.call_args[0]) == 1
        assert error.call_args[0][0].startswith(
            "Exception in shell extension 'extension2': custom exception\n")


def test_get_environment_variables():
    cmd = [
        'echo', 'FOO\nNAME=value\n\nSOMETHING\nNAME2=value with spaces']

    coroutine = get_environment_variables(cmd, shell=False)
    env = run_until_complete(coroutine)

    assert len(env.keys()) == 2
    assert 'NAME' in env.keys()
    assert env['NAME'] == 'value\nSOMETHING'
    assert 'NAME2' in env.keys()
    assert env['NAME2'] == 'value with spaces'

    # test with environment strings which isn't decodable
    async def check_output(cmd, **kwargs):
        return b'DECODE_ERROR=\x81\nNAME=value'
    with patch('colcon_core.shell.check_output', side_effect=check_output):
        with patch('colcon_core.shell.logger.warning') as warn:
            coroutine = get_environment_variables(['not-used'], shell=False)
            env = run_until_complete(coroutine)

    assert len(env.keys()) == 1
    assert 'NAME' in env.keys()
    assert env['NAME'] == 'value'
    # the raised decode error is catched and results in a warning message
    assert warn.call_count == 1
    assert len(warn.call_args[0]) == 1
    assert warn.call_args[0][0].startswith(
        "Failed to decode line from the environment using the encoding '")
    assert 'DECODE_ERROR=' in warn.call_args[0][0]


class Extension3(ShellExtensionPoint):
    PRIORITY = 105


class Extension4(ShellExtensionPoint):
    PRIORITY = 101


class Extension5(ShellExtensionPoint):
    PRIORITY = 110


def test_create_environment_hook():
    with EntryPointContext(extension1=Extension1, extension2=Extension2):
        # no primary shell extension
        with pytest.raises(RuntimeError) as e:
            create_environment_hook(None, None, None, None, None)
        assert str(e.value).endswith(
            'Could not find a primary shell extension for creating an '
            'environment hook')

    with EntryPointContext(
        extension3=Extension3, extension4=Extension4, extension5=Extension5
    ):
        extensions = get_shell_extensions()

        # one invalid, two valid return values
        extensions[105]['extension3'].create_hook_prepend_value = Mock()
        extensions[101]['extension4'].create_hook_prepend_value = Mock(
            return_value=Path('/some/path/sub/hookA'))
        extensions[110]['extension5'].create_hook_prepend_value = Mock(
            return_value=Path('/some/path/sub/hookB'))
        with patch('colcon_core.shell.logger.error') as error:
            hooks = create_environment_hook(None, None, None, None, None)
        assert len(hooks) == 2
        assert str(hooks[0]) == '/some/path/sub/hookB'.replace('/', os.sep)
        assert str(hooks[1]) == '/some/path/sub/hookA'.replace('/', os.sep)
        # the raised exception is catched and results in an error message
        assert error.call_count == 1
        assert len(error.call_args[0]) == 1
        assert error.call_args[0][0].startswith(
            "Exception in shell extension 'extension3': "
            'create_hook_prepend_value() should return a Path object')

        # invalid mode
        with pytest.raises(NotImplementedError):
            create_environment_hook(
                None, None, None, None, None, mode='invalid')


def test_get_colcon_prefix_path():
    # ignore deprecation warning
    with patch('colcon_core.shell.warnings.warn') as warn:
        # empty environment variable
        with EnvironmentContext(COLCON_PREFIX_PATH=''):
            prefix_path = get_colcon_prefix_path()
            assert prefix_path == []

        # extra path separator
        with EnvironmentContext(COLCON_PREFIX_PATH=os.pathsep):
            prefix_path = get_colcon_prefix_path(skip='/path/to/skip')
            assert prefix_path == []

        with TemporaryDirectory(prefix='test_colcon_') as basepath:
            basepath = Path(basepath)
            with EnvironmentContext(COLCON_PREFIX_PATH=os.pathsep.join(
                [str(basepath), str(basepath)]
            )):
                # multiple results
                prefix_path = get_colcon_prefix_path(skip='/path/to/skip')
                assert prefix_path == [str(basepath), str(basepath)]

                # skipping results
                prefix_path = get_colcon_prefix_path(skip=str(basepath))
                assert prefix_path == []

            # skipping non-existing results
            with EnvironmentContext(COLCON_PREFIX_PATH=os.pathsep.join(
                [str(basepath), str(basepath / 'non-existing-sub')]
            )):
                with patch('colcon_core.shell.logger.warning') as warn:
                    prefix_path = get_colcon_prefix_path()
                assert prefix_path == [str(basepath)]
                assert warn.call_count == 1
                assert len(warn.call_args[0]) == 1
                assert warn.call_args[0][0].endswith(
                    "non-existing-sub' in the environment variable "
                    "COLCON_PREFIX_PATH doesn't exist")
                # suppress duplicate warning
                with patch('colcon_core.shell.logger.warning') as warn:
                    prefix_path = get_colcon_prefix_path()
                assert prefix_path == [str(basepath)]
                assert warn.call_count == 0


def test_check_dependency_availability():
    with TemporaryDirectory(prefix='test_colcon_') as prefix_path:
        prefix_path = Path(prefix_path)

        dependencies = OrderedDict()
        dependencies['pkgA'] = prefix_path

        # missing package
        with pytest.raises(RuntimeError) as e:
            check_dependency_availability(
                dependencies, script_filename='package.ext')
        assert len(dependencies) == 1
        assert 'Failed to find the following files:' in str(e.value)
        assert str(prefix_path / 'share' / 'pkgA' / 'package.ext') \
            in str(e.value)
        assert 'Check that the following packages have been built:' \
            in str(e.value)
        assert '- pkgA' in str(e.value)

        # package in workspace
        (prefix_path / 'share' / 'pkgA').mkdir(parents=True)
        (prefix_path / 'share' / 'pkgA' / 'package.ext').write_text('')
        check_dependency_availability(
            dependencies, script_filename='package.ext')
        assert len(dependencies) == 1

        # package in environment
        dependencies['pkgA'] = prefix_path / 'invalid'
        with patch(
            'colcon_core.shell.find_installed_packages_in_environment',
            side_effect=lambda: {'pkgA': prefix_path / 'env'}
        ):
            with patch('colcon_core.shell.logger.warning') as warn:
                check_dependency_availability(
                    dependencies, script_filename='package.ext')
        assert len(dependencies) == 0
        assert warn.call_count == 1
        assert len(warn.call_args[0]) == 1
        assert warn.call_args[0][0].startswith(
            "The following packages are in the workspace but haven't been "
            'built:')
        assert '- pkgA' in warn.call_args[0][0]
        assert 'They are being used from the following locations instead:' \
            in warn.call_args[0][0]
        assert str(prefix_path / 'env') in warn.call_args[0][0]
        assert '--packages-ignore pkgA' in warn.call_args[0][0]


def test_find_installed_packages_in_environment():
    with TemporaryDirectory(prefix='test_colcon_') as prefix_path:
        prefix_path = Path(prefix_path)
        prefix_path1 = prefix_path / 'one'
        prefix_path2 = prefix_path / 'two'

        with patch(
            'colcon_core.shell.get_chained_prefix_path',
            return_value=[prefix_path1, prefix_path2]
        ):
            # not used prefixes result debug messages
            with patch('colcon_core.shell.logger.debug') as debug:
                find_installed_packages_in_environment()
            assert debug.call_count == 2

            # the package is picked up from the first prefix
            with patch(
                'colcon_core.shell.find_installed_packages',
                side_effect=lambda p: {'pkgA': p}
            ):
                packages = find_installed_packages_in_environment()
        assert len(packages) == 1
        assert 'pkgA' in packages
        assert packages['pkgA'] == prefix_path1


def test_find_installed_packages():
    with TemporaryDirectory(prefix='test_colcon_') as install_base:
        install_base = Path(install_base)

        # install base doesn't exist
        assert find_installed_packages(install_base) is None

        # unknown install layout
        marker_file = install_base / '.colcon_install_layout'
        marker_file.write_text('unknown')
        assert find_installed_packages(install_base) is None

        # package index directory doesn't exist
        marker_file.write_text('merged')
        packages = find_installed_packages(install_base)
        assert len(packages) == 0

        with patch(
            'colcon_core.shell.get_relative_package_index_path',
            return_value=Path('relative/package/index')
        ) as rel_path:
            # setup for isolated case
            (install_base / 'dummy_file').write_text('')
            (install_base / '.hidden_dir').mkdir()
            (install_base / 'dummy_dir' / rel_path() / 'dummy_dir').mkdir(
                parents=True)
            (install_base / 'pkgA' / rel_path()).mkdir(parents=True)
            (install_base / 'pkgA' / rel_path() / 'pkgA').write_text('')

            # setup for merged case
            (install_base / rel_path() / 'dummy_dir').mkdir(parents=True)
            (install_base / rel_path() / '.dummy').write_text('')
            (install_base / rel_path() / 'pkgB').write_text('')
            (install_base / rel_path() / 'pkgC').write_text('')

            marker_file.write_text('isolated')
            packages = find_installed_packages(install_base)
            assert len(packages) == 1
            assert 'pkgA' in packages.keys()
            assert packages['pkgA'] == install_base / 'pkgA'

            marker_file.write_text('merged')
            packages = find_installed_packages(install_base)
            assert len(packages) == 2
            assert 'pkgB' in packages.keys()
            assert packages['pkgC'] == install_base
            assert 'pkgC' in packages.keys()
            assert packages['pkgB'] == install_base



# Copyright 2016-2019 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from pathlib import Path
import sys

from colcon_core import shell
from colcon_core.plugin_system import satisfies_version
from colcon_core.plugin_system import SkipExtensionException
from colcon_core.prefix_path import get_chained_prefix_path
from colcon_core.shell import check_dependency_availability
from colcon_core.shell import get_environment_variables
from colcon_core.shell import logger
from colcon_core.shell import ShellExtensionPoint
from colcon_core.shell.template import expand_template


class BatShell(ShellExtensionPoint):
    """Generate `.bat` scripts to extend the environment."""

    # the priority needs to be higher than the default for primary shells
    PRIORITY = 200

    FORMAT_STR_COMMENT_LINE = ':: {comment}'
    FORMAT_STR_SET_ENV_VAR = 'set "{name}={value}"'
    FORMAT_STR_USE_ENV_VAR = '%{name}%'
    FORMAT_STR_INVOKE_SCRIPT = 'call:_colcon_prefix_bat_call_script ' \
        '"{script_path}"'
    # can't use `if` here since each line is being `call`-ed
    FORMAT_STR_REMOVE_TRAILING_SEPARATOR = \
        'call:_colcon_prefix_bat_strip_trailing_semicolon "{name}"'

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(ShellExtensionPoint.EXTENSION_POINT_VERSION, '^2.1')
        if sys.platform != 'win32' and not shell.use_all_shell_extensions:
            raise SkipExtensionException('Not used on non-Windows systems')

    def create_prefix_script(self, prefix_path, merge_install):  # noqa: D102
        prefix_env_path = prefix_path / 'local_setup.bat'
        logger.info("Creating prefix script '%s'" % prefix_env_path)
        expand_template(
            Path(__file__).parent / 'template' / 'prefix.bat.em',
            prefix_env_path,
            {
                'python_executable': sys.executable,
                'merge_install': merge_install,
                'package_script_no_ext': 'package',
            })

        prefix_util_path = prefix_path / '_local_setup_util_bat.py'
        logger.info("Creating prefix util module '%s'" % prefix_util_path)
        expand_template(
            self._get_prefix_util_template_path(),
            prefix_util_path, {'shell_extension': self})

        prefix_chain_env_path = prefix_path / 'setup.bat'
        logger.info(
            "Creating prefix chain script '%s'" % prefix_chain_env_path)
        expand_template(
            Path(__file__).parent / 'template' / 'prefix_chain.bat.em',
            prefix_chain_env_path,
            {
                'chained_prefix_path': get_chained_prefix_path(
                    skip=prefix_path),
                'prefix_script_no_ext': 'local_setup',
            })

    def create_package_script(
        self, prefix_path, pkg_name, hooks
    ):  # noqa: D102
        pkg_env_path = prefix_path / 'share' / pkg_name / 'package.bat'
        logger.info("Creating package script '%s'" % pkg_env_path)
        expand_template(
            Path(__file__).parent / 'template' / 'package.bat.em',
            pkg_env_path,
            {
                'hooks': list(filter(
                    lambda hook: str(hook[0]).endswith('.bat'), hooks)),
            })

    def create_hook_set_value(
        self, env_hook_name, prefix_path, pkg_name, name, value,
    ):  # noqa: D102
        hook_path = prefix_path / 'share' / pkg_name / 'hook' / \
            ('%s.bat' % env_hook_name)
        logger.info("Creating environment hook '%s'" % hook_path)
        if value == '':
            value = '%COLCON_PREFIX_PATH%'
        expand_template(
            Path(__file__).parent / 'template' / 'hook_set_value.bat.em',
            hook_path, {'name': name, 'value': value})
        return hook_path

    def create_hook_prepend_value(
        self, env_hook_name, prefix_path, pkg_name, name, subdirectory,
    ):  # noqa: D102
        hook_path = prefix_path / 'share' / pkg_name / 'hook' / \
            ('%s.bat' % env_hook_name)
        logger.info("Creating environment hook '%s'" % hook_path)
        expand_template(
            Path(__file__).parent / 'template' / 'hook_prepend_value.bat.em',
            hook_path,
            {
                'name': name,
                'subdirectory': subdirectory,
            })
        return hook_path

    async def generate_command_environment(
        self, task_name, build_base, dependencies,
    ):  # noqa: D102
        if sys.platform != 'win32':
            raise SkipExtensionException('Not usable on non-Windows systems')

        # check if all dependencies are available
        # removes dependencies available in the environment from the parameter
        check_dependency_availability(
            dependencies, script_filename='package.bat')

        hook_path = build_base / ('colcon_command_prefix_%s.bat' % task_name)
        expand_template(
            Path(__file__).parent / 'template' / 'command_prefix.bat.em',
            hook_path,
            {'dependencies': dependencies})

        cmd = [str(hook_path), '&&', 'set']
        env = await get_environment_variables(cmd, cwd=str(build_base))

        # write environment variables to file for debugging
        env_path = build_base / (
            'colcon_command_prefix_%s.bat.env' % task_name)
        with env_path.open('w') as h:
            for key in sorted(env.keys()):
                value = env[key]
                h.write('{key}={value}\n'.format_map(locals()))

        return env



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import os
import shutil
import sys
import traceback
import warnings

from colcon_core.event.command import Command
from colcon_core.event.command import CommandEnded
from colcon_core.event.job import JobProgress
from colcon_core.event.output import StderrLine
from colcon_core.event.output import StdoutLine
from colcon_core.logging import colcon_logger
from colcon_core.plugin_system import instantiate_extensions
from colcon_core.plugin_system import order_extensions_by_name
from colcon_core.subprocess import run as colcon_core_subprocess_run

logger = colcon_logger.getChild(__name__)


class TaskContext:
    """The context provided to tasks."""

    def __init__(self, *, pkg, args, dependencies):
        """
        Construct a TaskContext.

        :param pkg: The package descriptor
        :param args: The parsed command line arguments
        :param dependencies: The ordered dictionary mapping dependency names to
          their paths
        """
        self.event_queue = None
        self.pkg = pkg
        self.args = args
        self.dependencies = dependencies

    def put_event_into_queue(self, event):
        """
        Post a message event into the event queue.

        The method is not implemented by default and will be replaced at
        runtime once the event queue is known.

        :param event: The event
        """
        raise NotImplementedError()


class TaskExtensionPoint:
    """
    The interface for task extensions.

    A task extension performs a specific logic and/or command sequence.

    For each instance the attributes `TASK_NAME` and `PACKAGE_TYPE` are being
    set to the (parent) basename of the entry point registering the extension.
    """

    """The version of the task extension interface."""
    EXTENSION_POINT_VERSION = '1.0'

    def add_arguments(self, *, parser):
        """
        Add command line arguments specific to the task.

        The method is intended to be overridden in a subclass.

        :param parser: The argument parser
        """
        pass

    def set_context(self, *, context):
        """
        Set the context before the task is being `__call__`-ed.

        :param context: The task context
        """
        self.context = context

    async def __call__(self, *args, **kwargs):
        """
        Execute the task extension logic.

        This method relays the call to a method named `self.TASK_NAME` which
        should be implemented in subclasses.

        :returns: The return code
        """
        task_method = getattr(self, self.TASK_NAME)
        return await task_method(*args, **kwargs)

    def print(self, msg, *, file=None):  # noqa: A003
        """
        Post a message event into the event queue.

        :param msg: The message
        :param file: The sink to write the message to. An argument of `None` or
          `sys.stdout` posts a `StdoutLine` event, `sys.stderr` posts a
          `StderrLine` event.
        """
        if file is None or file == sys.stdout:
            data = StdoutLine(msg + '\n')
        elif file == sys.stderr:
            data = StderrLine(msg + '\n')
        else:
            assert False, 'Unknown file object: ' + str(file)
        self.context.put_event_into_queue(data)

    def progress(self, message):
        """
        Post a progress event into the event queue.

        :param msg: The message
        """
        self.context.put_event_into_queue(
            JobProgress(self.context.pkg.name, message))


async def check_call(
    context, cmd, *, cwd=None, env=None, shell=False, use_pty=None
):  # pragma: no cover
    """
    Run the command described by cmd.

    Post a `Command` event to the queue describing the exact invocation in
    order to allow reproducing it.
    All output to `stdout` and `stderr` is posted as `StdoutLine` and
    `StderrLine` events to the event queue.

    This function has been depreated, use ``colcon_core.task.run()`` instead.

    :param cmd: The command and its arguments
    :param cwd: the working directory for the subprocess
    :param env: a dictionary with environment variables
    :param shell: whether to use the shell as the program to execute
    :param use_pty: whether to use a pseudo terminal
    :returns: the result of the completed process
    :rtype subprocess.CompletedProcess
    """
    warnings.warn(
        'colcon_core.task.check_call() has been deprecated, use '
        'colcon_core.task.run() instead', stacklevel=2)
    return await run(
        context, cmd, cwd=cwd, env=env, shell=shell, use_pty=use_pty)


async def run(
    context, cmd, *, cwd=None, env=None, shell=False, use_pty=None
):
    """
    Run the command described by cmd.

    Post a `Command` event to the queue describing the exact invocation in
    order to allow reproducing it.
    All output to `stdout` and `stderr` is posted as `StdoutLine` and
    `StderrLine` events to the event queue.

    :param cmd: The command and its arguments
    :param cwd: the working directory for the subprocess
    :param env: a dictionary with environment variables
    :param shell: whether to use the shell as the program to execute
    :param use_pty: whether to use a pseudo terminal
    :returns: the result of the completed process
    :rtype subprocess.CompletedProcess
    """
    def stdout_callback(line):
        context.put_event_into_queue(StdoutLine(line))

    def stderr_callback(line):
        context.put_event_into_queue(StderrLine(line))

    context.put_event_into_queue(
        Command(cmd, cwd=cwd, env=env, shell=shell))
    completed = await colcon_core_subprocess_run(
        cmd, stdout_callback, stderr_callback,
        cwd=cwd, env=env, shell=shell, use_pty=use_pty)
    context.put_event_into_queue(
        CommandEnded(
            cmd, cwd=cwd, env=env, shell=shell,
            returncode=completed.returncode))
    return completed


def get_task_extensions(task_name, *, unique_instance=False):
    """
    Get the available task extensions.

    The extensions are ordered by their entry point name.

    :param str task_name: The entry point name identifying a group of task
      extensions
    :param bool unique_instance: The flag if the returned instances should be
      unique or cached instances can be returned instead
    :rtype: OrderedDict
    """
    extensions = instantiate_extensions(
        task_name, unique_instance=unique_instance)
    task_basename = task_name.split('.')[-1]
    for name in list(extensions.keys()):
        extension = extensions[name]
        assert hasattr(extension, task_basename)
        extension.TASK_NAME = task_basename
        extension.PACKAGE_TYPE = name
    return order_extensions_by_name(extensions)


def add_task_arguments(parser, task_name):
    """
    Add the command line arguments for the task extensions.

    :param parser: The argument parser
    :param str task_name: The entry point name identifying a group of task
      extensions
    """
    extensions = get_task_extensions(task_name, unique_instance=True)
    for extension_name, extension in extensions.items():
        group = parser.add_argument_group(
            title="Arguments for '{extension_name}' packages"
            .format_map(locals()))
        try:
            retval = extension.add_arguments(parser=group)
            assert retval is None, 'add_arguments() should return None'
        except Exception as e:  # noqa: F841
            # catch exceptions raised in task extension
            exc = traceback.format_exc()
            logger.error(
                'Exception in task extension '
                "'{extension.TASK_NAME}.{extension.PACKAGE_TYPE}': {e}\n{exc}"
                .format_map(locals()))
            # skip failing extension, continue with next one


def get_task_extension(task_name, package_type):
    """
    Get a specific task extension.

    :param str task_name: The entry point name identifying a group of task
      extensions
    :param str package_type: The package type identifying a task extension
      within the group
    :returns: The task extension
    """
    extensions = instantiate_extensions(
        task_name, unique_instance=True)
    if package_type not in extensions:
        return None
    extension = extensions[package_type]
    extension.TASK_NAME = task_name.split('.')[-1]
    extension.PACKAGE_TYPE = package_type
    return extension


def create_file(args, rel_path, *, content=None):
    """
    Create a file within the install base.

    Creates the containing directory if necessary.

    :param args: The parsed command line arguments containing the install base
    :param str rel_path: The relative path of file
    :param str content: The content of the created file
    :returns: The task extension
    """
    dst = os.path.join(args.install_base, rel_path)
    os.makedirs(os.path.dirname(dst), exist_ok=True)
    with open(dst, 'w') as h:
        if content is not None:
            h.write(content)


def install(args, rel_src, rel_dst):
    """
    Install or symlink a file.

    Creates the containing directory if necessary.

    :param args: The parsed command line arguments containing the source path
      as well as the install base
    :param str rel_src: The source path relative to the path
    :param str rel_dst: The destination path relative to the install base
    """
    src = os.path.join(args.path, rel_src)
    dst = os.path.join(args.install_base, rel_dst)
    os.makedirs(os.path.dirname(dst), exist_ok=True)
    if not args.symlink_install:
        if os.path.islink(dst):
            os.unlink(dst)
        shutil.copy(src, dst)
    else:
        if os.path.islink(dst):
            if not os.path.exists(dst) or not os.path.samefile(src, dst):
                os.unlink(dst)
        elif os.path.isfile(dst):
            os.remove(dst)
        elif os.path.isdir(dst):
            shutil.rmtree(dst)
        if not os.path.exists(dst):
            os.symlink(src, dst)



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from pathlib import Path
from tempfile import TemporaryDirectory

from colcon_core.environment.path import PathEnvironment
from mock import patch


def test_path():
    extension = PathEnvironment()

    with TemporaryDirectory(prefix='test_colcon_') as prefix_path:
        prefix_path = Path(prefix_path)
        with patch(
            'colcon_core.shell.create_environment_hook',
            return_value=['/some/hook', '/other/hook']
        ):
            # bin directory does not exist
            hooks = extension.create_environment_hooks(prefix_path, 'pkg_name')
            assert len(hooks) == 0

            # bin directory exists, but empty
            (prefix_path / 'bin').mkdir()
            hooks = extension.create_environment_hooks(prefix_path, 'pkg_name')
            assert len(hooks) == 0

            # bin directory exists, but only subdirectories
            (prefix_path / 'bin' / 'subdir').mkdir()
            hooks = extension.create_environment_hooks(prefix_path, 'pkg_name')
            assert len(hooks) == 0

            # bin directory exists, with file
            (prefix_path / 'bin' / 'hook').write_text('')
            hooks = extension.create_environment_hooks(prefix_path, 'pkg_name')
            assert len(hooks) == 2

            # bin directory exists, with files
            (prefix_path / 'bin' / 'hook2').write_text('')
            hooks = extension.create_environment_hooks(prefix_path, 'pkg_name')
            assert len(hooks) == 2



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from collections import OrderedDict
import traceback

from colcon_core.entry_point import load_entry_points
from colcon_core.logging import colcon_logger
from pkg_resources import parse_version

logger = colcon_logger.getChild(__name__)


class SkipExtensionException(Exception):
    """
    Exception to signal that an extension should be skipped.

    It should be raised in the constructor of an extension.
    """

    pass


def instantiate_extensions(
    group_name, *, exclude_names=None, unique_instance=False
):
    """
    Instantiate all extensions within a group.

    :param str group_name: the name of the `entry_point` group
    :param exclude_names: a list of entry point names to exclude
    :param bool unique_instance: The flag if each extensions should be
      instantiated even when it has been created and cached before
    :returns: dict of extensions
    """
    extension_types = load_entry_points(
        group_name, exclude_names=exclude_names)
    extension_instances = {}
    for extension_name, extension_class in extension_types.items():
        extension_instance = _instantiate_extension(
            group_name, extension_name, extension_class,
            unique_instance=unique_instance)
        if extension_instance is None:
            continue
        extension_instances[extension_name] = extension_instance
    return extension_instances


_extension_instances = {}


def _instantiate_extension(
    group_name, extension_name, extension_class, *, unique_instance=False
):
    global _extension_instances
    if not unique_instance and extension_class in _extension_instances:
        return _extension_instances[extension_class]

    try:
        extension_instance = extension_class()
        assert isinstance(extension_instance, object), \
            'invocation should return an object'
    except SkipExtensionException as e:  # noqa: F841
        logger.info(
            "Skipping extension '{group_name}.{extension_name}': {e}"
            .format_map(locals()))
        extension_instance = None
    except Exception as e:  # noqa: F841
        # catch exceptions raised in extension constructor
        exc = traceback.format_exc()
        logger.error(
            'Exception instantiating extension '
            "'{group_name}.{extension_name}': {e}\n{exc}".format_map(locals()))
        extension_instance = None
    if not unique_instance:
        _extension_instances[extension_class] = extension_instance
    return extension_instance


def order_extensions_by_name(extensions):
    """
    Order the extensions based on the entry point name.

    :param extensions: a dict mapping the entry point name to the
      extension instance
    :returns: ordered dict mapping the entry point name to the extension
      instance
    """
    return OrderedDict(
        sorted(extensions.items(), key=lambda pair: pair[0]))


def order_extensions_by_priority(extensions):
    """
    Order the extensions based on their `PRIORITY` attribute.

    :param extensions: a dict mapping the entry point name to the
      extension instance
    :returns: ordered dict mapping the entry point name to the extension
      instance
    """
    # use negative priority and inverted reverse flag
    # to get highest priorities first and
    # extensions withing the same priority group in alphabetical order
    return OrderedDict(
        sorted(
            extensions.items(),
            key=lambda pair: (-pair[1].PRIORITY, pair[0])))


def order_extensions_grouped_by_priority(extensions):
    """
    Group the extensions based on their `PRIORITY` attribute.

    Each priority group is ordered by the entry point name.

    :param extensions: a dict mapping the entry point name to the
      extension instance
    :returns: ordered dict mapping the priority to an ordered dict which maps
      the entry point name to the extension instance
    """
    # use negative priority and inverted reverse flag
    # to get highest priorities first and
    # extensions withing the same priority group in alphabetical order
    ordered_extensions = OrderedDict(
        sorted(
            extensions.items(),
            key=lambda pair: (-pair[1].PRIORITY, pair[0])))
    grouped_extensions = OrderedDict()
    for name, extension in ordered_extensions.items():
        priority = extension.PRIORITY
        if priority not in grouped_extensions:
            grouped_extensions[priority] = OrderedDict()
        grouped_extensions[priority][name] = extension
    return grouped_extensions


def get_first_line_doc(any_type):
    """
    Get the first line of the docstring of an object.

    Leading spaces are ignored and a trailing dot is being removed.

    :param any_type: the object
    :returns: the first line of the `__doc__` attribute of the object,
      otherwise an empty string
    :rtype: str
    """
    if not any_type.__doc__:
        return ''
    lines = list(filter(
        lambda line: line.lstrip(), any_type.__doc__.splitlines()))
    if not lines:
        return ''
    line = lines[0].lstrip()
    if line.endswith('.'):
        line = line[:-1]
    return line


def satisfies_version(version, caret_range):
    """
    Check if a version is within a caret range.

    :param str version: the version number to check (e.g. `1.2.3`)
    :param str caret_range: the caret range (e.g. `^1.2`)
    :raises RuntimeError: if the version doesn't match the caret range
    """
    assert caret_range.startswith('^'), 'Only supports caret ranges'
    extension_point_version = parse_version(version)
    extension_version = parse_version(caret_range[1:])
    next_extension_version = _get_upper_bound_caret_version(
        extension_version)

    if extension_point_version < extension_version:
        raise RuntimeError(
            'Extension point is too old (%s), the extension requires '
            "'%s'" % (extension_point_version, extension_version))

    if extension_point_version >= next_extension_version:
        raise RuntimeError(
            'Extension point is newer (%s), than what the extension '
            "supports '%s'" % (extension_point_version, extension_version))


def _get_upper_bound_caret_version(version):
    parts = version.base_version.split('.')
    if len(parts) < 2:
        parts += [0] * (2 - len(parts))
    major, minor = [int(p) for p in parts[:2]]
    if major > 0:
        major += 1
        minor = 0
    else:
        minor += 1
    return parse_version('%d.%d.0' % (major, minor))



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import sys
import time

from colcon_core.event.job import JobEnded
from colcon_core.event.job import JobStarted
from colcon_core.event.test import TestFailure
from colcon_core.event_handler import EventHandlerExtensionPoint
from colcon_core.event_handler import format_duration
from colcon_core.plugin_system import satisfies_version
from colcon_core.subprocess import SIGINT_RESULT


class ConsoleStartEndEventHandler(EventHandlerExtensionPoint):
    """
    Output task name on start/end.

    The extension handles events of the following types:
    - :py:class:`colcon_core.event.job.JobStarted`
    - :py:class:`colcon_core.event.job.JobEnded`
    - :py:class:`colcon_core.event.test.TestFailure`
    """

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            EventHandlerExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')
        self._start_times = {}
        self._with_test_failures = set()

    def __call__(self, event):  # noqa: D102
        data = event[0]

        if isinstance(data, JobStarted):
            print(
                'Starting >>> {data.identifier}'.format_map(locals()),
                flush=True)
            self._start_times[data.identifier] = time.monotonic()

        elif isinstance(data, TestFailure):
            job = event[1]
            self._with_test_failures.add(job)

        elif isinstance(data, JobEnded):
            duration = \
                time.monotonic() - self._start_times[data.identifier]
            duration_string = format_duration(duration)
            if not data.rc:
                msg = 'Finished <<< {data.identifier} [{duration_string}]' \
                    .format_map(locals())
                job = event[1]
                if job in self._with_test_failures:
                    msg += '\t[ with test failures ]'
                writable = sys.stdout

            elif data.rc == SIGINT_RESULT:
                msg = 'Aborted  <<< {data.identifier} [{duration_string}]' \
                    .format_map(locals())
                writable = sys.stdout

            else:
                msg = 'Failed   <<< {data.identifier} ' \
                    '[{duration_string}, exited with code {data.rc}]' \
                    .format_map(locals())
                writable = sys.stderr

            print(msg, file=writable, flush=True)



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from argparse import ArgumentParser

from colcon_core.argument_parser import ArgumentParserDecorator
from colcon_core.argument_parser import ArgumentParserDecoratorExtensionPoint
from colcon_core.argument_parser import decorate_argument_parser
from colcon_core.argument_parser import get_argument_parser_extensions
from mock import Mock
from mock import patch
import pytest

from .entry_point_context import EntryPointContext


class Extension1(ArgumentParserDecoratorExtensionPoint):
    PRIORITY = 80


class Extension2(ArgumentParserDecoratorExtensionPoint):
    pass


def test_get_argument_parser_extensions():
    with EntryPointContext(extension1=Extension1, extension2=Extension2):
        extensions = get_argument_parser_extensions()
        assert ['extension2', 'extension1'] == \
            list(extensions.keys())


def decorate_argument_parser_mock(*, parser):
    class Decorator():

        def __init__(self, parser):
            self.parser = parser

        def add_argument(self, *args, **kwargs):
            pass  # pragma: no cover
    return Decorator(parser)


def test_decorate_argument_parser():
    parser = ArgumentParser()
    with EntryPointContext(extension1=Extension1, extension2=Extension2):
        extensions = get_argument_parser_extensions()

        # one invalid return value, one not implemented
        extensions['extension1'].decorate_argument_parser = Mock(
            return_value=None)
        with patch('colcon_core.argument_parser.logger.error') as error:
            decorated_parser = decorate_argument_parser(parser)
        assert decorated_parser == parser
        # the raised exceptions are catched and result in error messages
        assert error.call_count == 2
        assert len(error.call_args_list[0][0]) == 1
        assert error.call_args_list[0][0][0].startswith(
            "Exception in argument parser decorator extension 'extension2': "
            '\n')
        assert error.call_args_list[0][0][0].endswith(
            '\nNotImplementedError\n')
        assert len(error.call_args_list[1][0]) == 1
        assert error.call_args_list[1][0][0].startswith(
            "Exception in argument parser decorator extension 'extension1': "
            'decorate_argument_parser() should return a parser like object\n')

        # one exception, one valid decorator
        extensions['extension2'].decorate_argument_parser = Mock(
            side_effect=RuntimeError('custom exception'))
        extensions['extension1'].decorate_argument_parser = Mock(
            side_effect=decorate_argument_parser_mock)
        with patch('colcon_core.argument_parser.logger.error') as error:
            decorated_parser = decorate_argument_parser(parser)
        assert decorated_parser.parser == parser
        # the raised exception is catched and results in an error message
        assert error.call_count == 1
        assert len(error.call_args[0]) == 1
        assert error.call_args[0][0].startswith(
            "Exception in argument parser decorator extension 'extension2': "
            'custom exception\n')


class Decorator(ArgumentParserDecorator):

    def __init__(self, parser, **kwargs):
        self.foo = 'foo'
        super().__init__(parser, **kwargs)


def test_argument_parser_decorator():
    parser = ArgumentParser()

    # __getattr__
    decorator = ArgumentParserDecorator(parser)
    assert decorator.add_argument == parser.add_argument

    del decorator.__dict__['_parser']
    with pytest.raises(AttributeError):
        decorator.add_argument

    # __setattr__
    decorator = Decorator(parser)
    decorator.foo = 'bar'
    assert 'foo' in decorator.__dict__
    assert decorator.__dict__['foo'] == 'bar'

    decorator.add_argument = True
    assert parser.add_argument is True

    assert 'bar' not in decorator.__dict__
    del decorator.__dict__['_parser']
    decorator.bar = 'baz'
    assert 'bar' in decorator.__dict__
    assert decorator.__dict__['bar'] == 'baz'

    # nesting
    parser = ArgumentParser()
    decorator = Decorator(parser)
    group = decorator.add_argument_group()
    group.add_argument('arg1')

    group = decorator.add_mutually_exclusive_group()
    group.add_argument('--arg2', action='store_true')

    group = decorator.add_subparsers(dest='verb')
    group = group.add_parser('do')
    group.add_argument('arg3')

    args = parser.parse_args(['ARG1', '--arg2', 'do', 'ARG3'])
    assert args.arg1 == 'ARG1'
    assert args.arg2 is True
    assert args.verb == 'do'
    assert args.arg3 == 'ARG3'



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import traceback
import types
import warnings

from colcon_core.logging import colcon_logger
from colcon_core.plugin_system import instantiate_extensions
from colcon_core.plugin_system import order_extensions_by_priority

logger = colcon_logger.getChild(__name__)


class ArgumentParserDecoratorExtensionPoint:
    """
    The interface for argument parser decorator extensions.

    An argument parser decorator extension performs additional functionality
    when adding command line arguments.

    For each instance the attribute `ARGUMENT_PARSER_DECORATOR_NAME` is being
    set to the basename of the entry point registering the extension.
    """

    """The version of the argument parser decorator extension interface."""
    EXTENSION_POINT_VERSION = '1.0'

    """The default priority of argument parser decorator extensions."""
    PRIORITY = 100

    def decorate_argument_parser(self, *, parser):
        """
        Decorate an argument parser to perform additional functionality.

        This method must be overridden in a subclass.

        :param parser: The argument parser
        :returns: A decorator
        """
        raise NotImplementedError()


def get_argument_parser_extensions():
    """
    Get the available argument parser extensions.

    The extensions are ordered by their priority and entry point name.

    :rtype: OrderedDict
    """
    extensions = instantiate_extensions(__name__)
    for name, extension in extensions.items():
        extension.ARGUMENT_PARSER_DECORATOR_NAME = name
    return order_extensions_by_priority(extensions)


def decorate_argument_parser(parser):
    """
    Decorate the parser using the available argument parser extensions.

    :param parser: The argument parser
    :returns: The decorated parser
    """
    extensions = get_argument_parser_extensions()
    for extension in extensions.values():
        logger.log(
            1, 'decorate_argument_parser() %s',
            extension.ARGUMENT_PARSER_DECORATOR_NAME)
        try:
            decorated_parser = extension.decorate_argument_parser(
                parser=parser)
            assert hasattr(decorated_parser, 'add_argument'), \
                'decorate_argument_parser() should return a parser like object'
        except Exception as e:  # noqa: F841
            # catch exceptions raised in decorator extension
            exc = traceback.format_exc()
            logger.error(
                'Exception in argument parser decorator extension '
                "'{extension.ARGUMENT_PARSER_DECORATOR_NAME}': {e}\n{exc}"
                .format_map(locals()))
            # skip failing extension, continue with next one
        else:
            parser = decorated_parser

    return parser


class ArgumentParserDecorator:
    """
    Decorate an argument parser as well as all recursive subparsers.

    The methods and arguments are the same as :class:`argparse.ArgumentParser`.

    Subclasses can perform any kind of task when e.g. arguments are being added
    without being concerned in which part of the hierarchy it is added.

    Subclasses should not set any member variables directly but pass them as
    keyword arguments to the constructor.
    """

    def __init__(self, parser, **kwargs):
        """
        Decorate an argument parser.

        :param parser: The argument parser to decorate
        :param **kwargs: The keyword arguments are set as attributes on this
          instance
        """
        assert '_parser' not in kwargs
        kwargs['_parser'] = parser
        assert '_nested_decorators_' not in kwargs
        kwargs['_nested_decorators_'] = []
        assert '_group_decorators' not in kwargs
        kwargs['_group_decorators'] = []
        assert '_recursive_decorators' not in kwargs
        kwargs['_recursive_decorators'] = []
        for k, v in kwargs.items():
            self.__dict__[k] = v

    @property
    def _nested_decorators(self):  # pragma: no cover
        warnings.warn(
            'colcon_core.argument_parser.ArgumentParserDecorator.'
            '_nested_decorators is a private variable and has been '
            'deprecated', stacklevel=2)
        return self._nested_decorators_

    def __getattr__(self, name):
        """
        Get an attribute from this decorator if it exists or the decoree.

        :param str name: The name of the attribute
        :returns: The attribute value
        :raises AttributeError: if the attribute doesn't exist in either of the
          two instances
        """
        # when argcomplete changes self.__class__ at runtime
        # the instance might not have a _parser attribute anymore
        if '_parser' not in self.__dict__:
            raise AttributeError(name)
        # get attribute from decoree
        return getattr(self.__dict__['_parser'], name)

    def __setattr__(self, name, value):
        """
        Set an attribute value on this decorator if it exists or the decoree.

        :param str name: The name of the attribute
        :param value: The attribute value
        """
        # overwrite existing attribute
        if name in self.__dict__:
            self.__dict__[name] = value
            return
        # when argcomplete changes self.__class__ at runtime
        # the instance might not have a _parser attribute anymore
        if '_parser' not in self.__dict__:
            self.__dict__[name] = value
            return
        # get attribute on decoree
        setattr(self.__dict__['_parser'], name, value)

    def add_argument_group(self, *args, **kwargs):
        """
        Decorate group parser before adding.

        See :class:`argparse.ArgumentParser.add_argument_group` for the method
        arguments and return value.
        """
        group = self.__class__(
            self._parser.add_argument_group(*args, **kwargs))
        self._nested_decorators_.append(group)
        self._group_decorators.append(group)
        return group

    def add_mutually_exclusive_group(self, *args, **kwargs):
        """
        Decorate mutually exclusive group parser before adding.

        See :class:`argparse.ArgumentParser.add_mutually_exclusive_group` for
        the method arguments and return value.
        """
        group = self.__class__(
            self._parser.add_mutually_exclusive_group(*args, **kwargs))
        self._nested_decorators_.append(group)
        self._group_decorators.append(group)
        return group

    def add_parser(self, *args, **kwargs):
        """
        Decorate parser before adding.

        See :class:`argparse.ArgumentParser.add_parser` for the method
        arguments and return value.
        """
        parser = self.__class__(
            self._parser.add_parser(*args, **kwargs))
        self._nested_decorators_.append(parser)
        self._recursive_decorators.append(parser)
        return parser

    def add_subparsers(self, *args, **kwargs):
        """
        Decorate subparser before adding.

        See :class:`argparse.ArgumentParser.add_subparsers` for the method
        arguments and return value.
        """
        subparser = self.__class__(
            self._parser.add_subparsers(*args, **kwargs))
        self._nested_decorators_.append(subparser)
        self._recursive_decorators.append(subparser)
        return subparser


class SuppressUsageOutput:
    """Context manager to suppress help action during `parse_known_args`."""

    def __init__(self, parsers):
        """
        Construct a SuppressUsageOutput.

        :param parsers: The parsers
        """
        self._parsers = parsers
        self._callbacks = {}

    def __enter__(self):  # noqa: D105
        for p in self._parsers:
            self._callbacks[p] = p.print_help, p.exit
            # temporary prevent printing usage early if help is requested
            p.print_help = lambda: None
            # temporary prevent help action to exit early,
            # but keep exiting on invalid arguments
            p.exit = types.MethodType(_ignore_zero_exit(p.exit), p)

        return self

    def __exit__(self, *args):  # noqa: D105
        for p, callbacks in self._callbacks.items():
            p.print_help, p.exit = callbacks


def _ignore_zero_exit(original_exit_handler):
    def exit_(self, status=0, message=None):
        nonlocal original_exit_handler
        if status == 0:
            return
        return original_exit_handler(status=status, message=message)

    return exit_



# Copyright 2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import os
from pathlib import Path
from tempfile import TemporaryDirectory

from colcon_core.location import get_relative_package_index_path
from colcon_core.shell.template.prefix_util import get_packages
from colcon_core.shell.template.prefix_util import main
from colcon_core.shell.template.prefix_util import order_packages
from colcon_core.shell.template.prefix_util import reduce_cycle_set
from mock import patch
import pytest


def test_main(capsys):
    with patch(
        'colcon_core.shell.template.prefix_util.get_packages',
        return_value={'pkgA': set()}
    ):
        main([])
    out, err = capsys.readouterr()
    assert out == 'pkgA\n'
    assert not err


def test_get_packages():
    with TemporaryDirectory(prefix='test_colcon_') as prefix_path:
        prefix_path = Path(prefix_path)

        # check no packages in not merged install layout
        packages = get_packages(prefix_path, False)
        assert packages == {}

        # mock packages in not merged install layout
        subdirectory = get_relative_package_index_path()
        for pkg_name in ('pkgA', 'pkgB'):
            (prefix_path / pkg_name / subdirectory).mkdir(parents=True)
            (prefix_path / pkg_name / subdirectory / pkg_name).write_text(
                'depX')
        (prefix_path / 'dummy_dir').mkdir()
        (prefix_path / '.hidden_dir').mkdir()
        (prefix_path / 'dummy_file').write_text('')

        # check no packages in merged install layout
        packages = get_packages(prefix_path, True)
        assert packages == {}

        # mock packages in merged install layout
        (prefix_path / subdirectory).mkdir(parents=True)
        (prefix_path / subdirectory / 'pkgB').write_text('')
        (prefix_path / subdirectory / 'pkgC').write_text(
            os.pathsep.join(('pkgB', 'depC')))
        (prefix_path / subdirectory / 'dummy_dir').mkdir()
        (prefix_path / subdirectory / '.hidden_file').write_text('')

        # check packages and dependencies in not merged install layout
        packages = get_packages(prefix_path, False)
        assert len(packages) == 2
        assert 'pkgA' in packages
        assert packages['pkgA'] == set()
        assert 'pkgB' in packages
        assert packages['pkgB'] == set()

        # check packages and dependencies in not merged install layout
        packages = get_packages(prefix_path, True)
        assert len(packages) == 2
        assert 'pkgB' in packages
        assert packages['pkgB'] == set()
        assert 'pkgC' in packages
        assert packages['pkgC'] == {'pkgB'}


def test_order_packages():
    packages = {
        'pkgA': {'pkgC'},
        'pkgB': {},
        'pkgC': {},
    }
    ordered = order_packages(packages)
    assert ordered == ['pkgB', 'pkgC', 'pkgA']

    packages = {
        'pkgA': {'pkgB'},
        'pkgB': {'pkgA'},
        'pkgC': set(),
    }
    with pytest.raises(RuntimeError) as e:
        ordered = order_packages(packages)
    assert 'Circular dependency between:' in str(e.value)
    assert 'pkgA' in str(e.value)
    assert 'pkgB' in str(e.value)
    assert 'pkgC' not in str(e.value)


def test_reduce_cycle_set():
    packages = {
        'pkgA': {'pkgB'},
        'pkgB': set(),
    }
    reduce_cycle_set(packages)
    assert len(packages) == 0



# Copyright 2019 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import os

from colcon_core.logging import colcon_logger
from colcon_core.plugin_system import satisfies_version
from colcon_core.prefix_path import PrefixPathExtensionPoint

logger = colcon_logger.getChild(__name__)

_get_colcon_prefix_path_warnings = set()


class ColconPrefixPath(PrefixPathExtensionPoint):
    """Prefix path defined in the `COLCON_PREFIX_PATH` environment variable."""

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            PrefixPathExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')

    def extend_prefix_path(self, paths):  # noqa: D102
        global _get_colcon_prefix_path_warnings
        colcon_prefix_path = os.environ.get('COLCON_PREFIX_PATH', '')
        for path in colcon_prefix_path.split(os.pathsep):
            if not path:
                continue
            if not os.path.exists(path):
                if path not in _get_colcon_prefix_path_warnings:
                    logger.warning(
                        "The path '{path}' in the environment variable "
                        "COLCON_PREFIX_PATH doesn't exist"
                        .format_map(locals()))
                _get_colcon_prefix_path_warnings.add(path)
                continue
            paths.append(path)



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import asyncio
import sys

from colcon_core.subprocess import check_output
from colcon_core.subprocess import new_event_loop
from colcon_core.subprocess import run
import pytest

from .run_until_complete import run_until_complete


# TODO figure out how to avoid the stderr output
@pytest.mark.skip(
    reason='Results in stderr output due to a UnicodeDecodeError for the '
           'generated coverage files')
def test_check_output():
    coroutine = check_output(
        [sys.executable, '-c', r"print('line1\nline2')"], shell=True)
    output = run_until_complete(coroutine)
    assert output.decode().splitlines() == ['line1', 'line2']


# TODO figure out how to avoid the stderr output
@pytest.mark.skip(
    reason='Results in stderr output due to a UnicodeDecodeError for the '
           'generated coverage files')
def test_run():
    # without callbacks
    coroutine = run(
        [sys.executable, '-c', r"print('line1\nline2')"], None, None)
    completed_process = run_until_complete(coroutine)
    assert completed_process.returncode == 0

    # without callbacks, with pty
    coroutine = run(
        [sys.executable, '-c', r"print('line1\nline2')"], None, None,
        use_pty=True)
    completed_process = run_until_complete(coroutine)
    assert completed_process.returncode == 0

    # with callbacks
    stdout_lines = []
    stderr_lines = []

    def stdout_callback(line):
        nonlocal stdout_lines
        stdout_lines.append(line)

    def stderr_callback(line):
        nonlocal stderr_lines
        stderr_lines.append(line)

    coroutine = run(
        [sys.executable, '-c', r"print('line1\nline2')"],
        stdout_callback, stderr_callback)
    completed_process = run_until_complete(coroutine)
    assert completed_process.returncode == 0
    assert stdout_lines == [b'line1\n', b'line2\n']

    # with callbacks, with pty
    stdout_lines = []
    stderr_lines = []
    coroutine = run(
        [sys.executable, '-c', r"print('line1\nline2')"],
        stdout_callback, stderr_callback, use_pty=True)
    completed_process = run_until_complete(coroutine)
    assert completed_process.returncode == 0
    assert stdout_lines == [b'line1\n', b'line2\n']


# TODO figure out why no coverage is being generated
@pytest.mark.skip(
    reason='No coverage is being generated for this test')
def test_run_cancel():
    # with callbacks, canceled
    stdout_lines = []

    def stdout_callback(line):
        nonlocal stdout_lines
        stdout_lines.append(line)

    coroutine = run(
        [
            sys.executable, '-c',
            r"import time; time.sleep(1); print('line1\nline2')"],
        stdout_callback, None)
    loop = new_event_loop()
    asyncio.set_event_loop(loop)
    task = asyncio.Task(coroutine, loop=loop)
    assert task.cancel() is True
    try:
        with pytest.raises(asyncio.CancelledError):
            loop.run_until_complete(task)
    finally:
        loop.close()



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from sys import executable

from colcon_core.plugin_system import satisfies_version
from colcon_core.task import run
from colcon_core.task.python.test import PythonTestingStepExtensionPoint
from colcon_core.verb.test import logger


class SetuppyPythonTestingStep(PythonTestingStepExtensionPoint):
    """Use `unittest` to test packages."""

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            PythonTestingStepExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')

    def add_arguments(self, *, parser):  # noqa: D102
        parser.add_argument(
            '--unittest-args',
            nargs='*', metavar='*', type=str.lstrip,
            help='Pass arguments to Python unittests. '
            'Arguments matching other options must be prefixed by a space,\n'
            'e.g. --unittest-args " --help" (stdout might not be shown by '
            'default, e.g. add `--event-handlers console_cohesion+`)')

    def match(self, context, env, setup_py_data):  # noqa: D102
        return True

    async def step(self, context, env, setup_py_data):  # noqa: D102
        if context.args.retest_until_fail:
            logger.warning(
                "Ignored '--retest-until-fail' for package "
                "'{context.pkg.name}' since 'unittest' does not support the "
                'usage'.format_map(locals()))

        if context.args.retest_until_pass:
            logger.warning(
                "Ignored '--retest-until-pass' for package "
                "'{context.pkg.name}' since 'unittest' does not support the "
                'usage'.format_map(locals()))

        cmd = [executable, '-m', 'unittest', '-v']
        if context.args.unittest_args is not None:
            cmd += context.args.unittest_args
        completed = await run(
            context, cmd, cwd=context.args.path, env=env)
        return completed.returncode



# Copyright 2019 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import os
from pathlib import Path
from tempfile import TemporaryDirectory

from colcon_core.prefix_path import get_chained_prefix_path
from colcon_core.prefix_path import get_prefix_path_extensions
from colcon_core.prefix_path import PrefixPathExtensionPoint
from colcon_core.prefix_path.colcon import ColconPrefixPath
from mock import Mock
from mock import patch
import pytest

from .entry_point_context import EntryPointContext
from .environment_context import EnvironmentContext


class Extension1(PrefixPathExtensionPoint):
    PRIORITY = 90


class Extension2(PrefixPathExtensionPoint):
    pass


def test_extension_interface():
    extension = Extension1()
    with pytest.raises(NotImplementedError):
        extension.extend_prefix_path(None)


def test_get_prefix_path_extensions():
    with EntryPointContext(extension1=Extension1, extension2=Extension2):
        extensions = get_prefix_path_extensions()
    assert list(extensions.keys()) == [100, 90]
    assert list(extensions[100].keys()) == ['extension2']
    assert list(extensions[90].keys()) == ['extension1']


def test_get_chained_prefix_path():
    # empty environment variable
    with EnvironmentContext(COLCON_PREFIX_PATH=''):
        prefix_path = get_chained_prefix_path()
        assert prefix_path == []

    # extra path separator
    with EnvironmentContext(COLCON_PREFIX_PATH=os.pathsep):
        prefix_path = get_chained_prefix_path(skip='/path/to/skip')
        assert prefix_path == []

    ColconPrefixPath.PREFIX_PATH_NAME = 'colcon'
    with patch(
        'colcon_core.prefix_path.get_prefix_path_extensions',
        return_value={
            100: {'colcon': ColconPrefixPath()},
        }
    ):
        with TemporaryDirectory(prefix='test_colcon_') as basepath:
            basepath = Path(basepath)
            with EnvironmentContext(COLCON_PREFIX_PATH=os.pathsep.join(
                [str(basepath), str(basepath)]
            )):
                # multiple results, duplicates being skipped
                prefix_path = get_chained_prefix_path(skip='/path/to/skip')
                assert prefix_path == [str(basepath)]

                # skipping results
                prefix_path = get_chained_prefix_path(skip=str(basepath))
                assert prefix_path == []

            # skipping non-existing results
            with EnvironmentContext(COLCON_PREFIX_PATH=os.pathsep.join(
                [str(basepath), str(basepath / 'non-existing-sub')]
            )):
                with patch(
                    'colcon_core.prefix_path.colcon.logger.warning'
                ) as warn:
                    prefix_path = get_chained_prefix_path()
                assert prefix_path == [str(basepath)]
                assert warn.call_count == 1
                assert len(warn.call_args[0]) == 1
                assert warn.call_args[0][0].endswith(
                    "non-existing-sub' in the environment variable "
                    "COLCON_PREFIX_PATH doesn't exist")
                # suppress duplicate warning
                with patch(
                    'colcon_core.prefix_path.colcon.logger.warning'
                ) as warn:
                    prefix_path = get_chained_prefix_path()
                assert prefix_path == [str(basepath)]
                assert warn.call_count == 0

    with EntryPointContext(extension1=Extension1, extension2=Extension2):
        extensions = get_prefix_path_extensions()

        # one invalid return value, one not implemented
        extensions[100]['extension2'].extend_prefix_path = Mock(
            return_value=False)
        extensions[90]['extension1'].extend_prefix_path = Mock(
            return_value=None)
        with patch('colcon_core.prefix_path.logger.error') as error:
            get_chained_prefix_path()
        # the raised exception is catched and results in an error message
        assert error.call_count == 1
        assert len(error.call_args_list[0][0]) == 1
        assert error.call_args_list[0][0][0].startswith(
            "Exception in prefix path extension 'extension2': "
            'extend_prefix_path() should return None\n')



# Copyright 2016-2019 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from colcon_core.dependency_descriptor import DependencyDescriptor
from colcon_core.package_augmentation import logger
from colcon_core.package_augmentation \
    import PackageAugmentationExtensionPoint
from colcon_core.package_identification.python import get_configuration
from colcon_core.package_identification.python import is_reading_cfg_sufficient
from colcon_core.plugin_system import satisfies_version
from distlib.util import parse_requirement
from distlib.version import NormalizedVersion


class PythonPackageAugmentation(PackageAugmentationExtensionPoint):
    """
    Augment Python packages with information from `setup.cfg` files.

    Only packages which pass no arguments (or only a ``cmdclass``) to the
    ``setup()`` function in their ``setup.py`` file are being considered.
    """

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            PackageAugmentationExtensionPoint.EXTENSION_POINT_VERSION,
            '^1.0')

    def augment_package(
        self, desc, *, additional_argument_names=None
    ):  # noqa: D102
        if desc.type != 'python':
            return

        setup_py = desc.path / 'setup.py'
        if not setup_py.is_file():
            return

        setup_cfg = desc.path / 'setup.cfg'
        if not setup_cfg.is_file():
            return

        if not is_reading_cfg_sufficient(setup_py):
            return

        config = get_configuration(setup_cfg)

        version = config.get('metadata', {}).get('version')
        desc.metadata['version'] = version

        options = config.get('options', {})
        dependencies = extract_dependencies(options)
        for k, v in dependencies.items():
            desc.dependencies[k] |= v

        def getter(env):
            nonlocal options
            return options

        desc.metadata['get_python_setup_options'] = getter


def extract_dependencies(options):
    """
    Get the dependencies of the package.

    :param options: The dictionary from the options section of the setup.cfg
      file
    :returns: The dependencies
    :rtype: dict(string, set(DependencyDescriptor))
    """
    mapping = {
        'setup_requires': 'build',
        'install_requires': 'run',
        'tests_require': 'test',
    }
    dependencies = {}
    for option_name, dependency_type in mapping.items():
        dependencies[dependency_type] = set()
        for dep in options.get(option_name, []):
            dependencies[dependency_type].add(
                create_dependency_descriptor(dep))
    return dependencies


def create_dependency_descriptor(requirement_string):
    """
    Create a DependencyDescriptor from a PEP440 compliant string.

    See https://www.python.org/dev/peps/pep-0440/#version-specifiers

    :param str requirement_string: a PEP440 compliant requirement string
    :return: A descriptor with version constraints from the requirement string
    :rtype: DependencyDescriptor
    """
    symbol_mapping = {
        '==': 'version_eq',
        '!=': 'version_neq',
        '<=': 'version_lte',
        '>=': 'version_gte',
        '>': 'version_gt',
        '<': 'version_lt',
    }

    requirement = parse_requirement(requirement_string)
    metadata = {}
    for symbol, version in (requirement.constraints or []):
        if symbol in symbol_mapping:
            metadata[symbol_mapping[symbol]] = version
        elif symbol == '~=':
            metadata['version_gte'] = version
            metadata['version_lt'] = _next_incompatible_version(version)
        else:
            logger.warning(
                "Ignoring unknown symbol '{symbol}' in '{requirement}'"
                .format_map(locals()))
    return DependencyDescriptor(requirement.name, metadata=metadata)


def _next_incompatible_version(version):
    """
    Find the next non-compatible version.

    This is for use with the ~= compatible syntax. It will provide
    the first version that this version must be less than in order
    to be compatible.

    :param str version: PEP 440 compliant version number
    :return: The first version after this version that is not compatible
    :rtype: str
    """
    normalized = NormalizedVersion(version)
    parse_tuple = normalized.parse(version)
    version_tuple = parse_tuple[1]

    *unchanged, increment, dropped = version_tuple
    incremented = increment + 1

    version = unchanged
    version.append(incremented)
    # versions have a minimum length of 2
    if len(version) == 1:
        version.append(0)
    return '.'.join(map(str, version))



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import os
from pathlib import Path
from tempfile import TemporaryDirectory

from colcon_core.package_descriptor import PackageDescriptor
from colcon_core.package_discovery.path import _expand_wildcards
from colcon_core.package_discovery.path import PathPackageDiscovery
from colcon_core.package_identification import IgnoreLocationException
from mock import Mock
from mock import patch


def test_path_package_discovery():
    extension = PathPackageDiscovery()
    assert extension.has_default() is True


def test_add_arguments():
    extension = PathPackageDiscovery()
    parser = Mock()
    parser.add_argument = Mock()
    extension.add_arguments(parser=parser, with_default=True)
    assert parser.add_argument.call_count == 1


def test_has_parameters():
    extension = PathPackageDiscovery()
    args = Mock()
    args.paths = []
    assert extension.has_parameters(args=args) is False
    args.paths = ['/some/path']
    assert extension.has_parameters(args=args) is True


def identify(_, path):
    if path == '/empty/path':
        return None
    if path == '/skip/path':
        raise IgnoreLocationException()
    return PackageDescriptor(path)


def test_discover():
    extension = PathPackageDiscovery()
    args = Mock()
    args.paths = None
    assert extension.discover(args=args, identification_extensions={}) == set()

    args.paths = [
        '/empty/path',
        '/skip/path',
        '/same/path', '/same/path/../path',
        '/other/path']
    with patch(
        'colcon_core.package_discovery.path.identify', side_effect=identify
    ):
        descs = extension.discover(args=args, identification_extensions={})
        assert descs == {
            PackageDescriptor(os.path.realpath('/same/path')),
            PackageDescriptor(os.path.realpath('/other/path'))}


def test__expand_wildcards():
    with TemporaryDirectory(prefix='test_colcon_') as prefix_path:
        prefix_path = Path(prefix_path)
        (prefix_path / 'one').mkdir()
        (prefix_path / 'two').mkdir()
        (prefix_path / 'three').touch()

        paths = [
            '/some/path',
            str(prefix_path / '*')
        ]
        _expand_wildcards(paths)
        assert len(paths) == 3
        assert paths[0] == '/some/path'
        assert paths[1] == str((prefix_path / 'one'))
        assert paths[2] == str((prefix_path / 'two'))



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from collections.abc import Iterable
import os
from pathlib import Path
import traceback

from colcon_core.location import get_relative_package_index_path
from colcon_core.logging import colcon_logger
from colcon_core.plugin_system import instantiate_extensions
from colcon_core.plugin_system import order_extensions_by_priority
from colcon_core.shell import get_shell_extensions

logger = colcon_logger.getChild(__name__)


class EnvironmentExtensionPoint:
    """
    The interface for environment extensions.

    An environment extension creates environment hooks for a specific
    environment variable and uses the shell extensions to generate scripts for
    each supported shell.

    For each instance the attribute `ENVIRONMENT_NAME` is being set to the
    basename of the entry point registering the extension.
    """

    """The version of the environment extension interface."""
    EXTENSION_POINT_VERSION = '1.0'

    """The default priority of environment extensions."""
    PRIORITY = 100

    def create_environment_hooks(self, prefix_path, pkg_name):
        """
        Create the environment hooks for a package.

        This method must be overridden in a subclass.

        :param prefix_path: The prefix path of the package
        :param pkg_name: The package name
        :returns: iterable of generated hook paths
        :rtype: Iterable
        """
        raise NotImplementedError()


def get_environment_extensions():
    """
    Get the available environment extensions.

    The extensions are ordered by their priority and entry point name.

    :rtype: OrderedDict
    """
    extensions = instantiate_extensions(__name__)
    for name in list(extensions.keys()):
        extension = extensions[name]
        extension.ENVIRONMENT_NAME = name
    return order_extensions_by_priority(extensions)


def create_environment_scripts(
    pkg, args, *, default_hooks=None, additional_hooks=None
):
    """
    Create the environment scripts for a package.

    Also create a file with the runtime dependencies of each packages which can
    be used by the prefix scripts to source all packages in topological order.

    :param pkg: The package descriptor
    :param args: The parsed command line arguments
    :param list default_hooks: If none are parsed explicitly the hooks provided
      by :function:`create_environment_hooks` are used
    :param list additional_hooks: Any additional hooks which should be
      referenced by the generated scripts
    """
    prefix_path = Path(args.install_base)
    create_environment_scripts_only(
        prefix_path, pkg, default_hooks=default_hooks,
        additional_hooks=additional_hooks)
    create_file_with_runtime_dependencies(prefix_path, pkg)


def create_environment_scripts_only(
    prefix_path, pkg, *, default_hooks=None, additional_hooks=None
):
    """
    Create the environment scripts for a package.

    :param prefix_path: The prefix path
    :param pkg: The package descriptor
    :param list default_hooks: If none are parsed explicitly the hooks provided
      by :function:`create_environment_hooks` are used
    :param list additional_hooks: Any additional hooks which should be
      referenced by the generated scripts
    """
    logger.log(1, 'create_environment_scripts_only(%s)', pkg.name)

    hooks = []
    if default_hooks is None:
        default_hooks = create_environment_hooks(prefix_path, pkg.name)
    hooks += default_hooks
    if additional_hooks:
        hooks += additional_hooks
    hooks += pkg.hooks

    # ensure each hook is presented by a tuple
    # with the first element being a relative path
    # and the second element being a list of arguments
    hook_tuples = []
    for hook in hooks:
        hook_args = []
        if isinstance(hook, list) or isinstance(hook, tuple):
            hook_args = hook[1:]
            hook = hook[0]

        if os.path.isabs(str(hook)):
            hook = os.path.relpath(str(hook), start=str(prefix_path))

        hook_tuples.append((hook, hook_args))

    extensions = get_shell_extensions()
    for priority in extensions.keys():
        extensions_same_prio = extensions[priority]
        for extension in extensions_same_prio.values():
            try:
                retval = extension.create_package_script(
                    prefix_path, pkg.name, hook_tuples)
                assert retval is None, \
                    'create_package_script() should return None'
            except Exception as e:  # noqa: F841
                # catch exceptions raised in shell extension
                exc = traceback.format_exc()
                logger.error(
                    "Exception in shell extension '{extension.SHELL_NAME}': "
                    '{e}\n{exc}'.format_map(locals()))
                # skip failing extension, continue with next one


def create_file_with_runtime_dependencies(prefix_path, pkg):
    """
    Create a file with the runtime dependencies of the package.

    It can be used by the prefix scripts to source all packages in topological
    order.

    :param prefix_path: The prefix path
    :param pkg: The package descriptor
    """
    path = prefix_path / get_relative_package_index_path() / pkg.name
    logger.log(1, 'create_file_with_runtime_dependencies(%s)', path)
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(
        os.pathsep.join(sorted(pkg.dependencies.get('run', set()))))


def create_environment_hooks(prefix_path, pkg_name):
    """
    Create the environment hooks for a package.

    :param prefix_path: The prefix path of the package
    :param pkg_name: The package name
    :returns: iterable of generated hook paths
    :rtype: Iterable
    """
    prefix_path = Path(prefix_path)
    all_hooks = []
    extensions = get_environment_extensions()
    for extension in extensions.values():
        try:
            hooks = extension.create_environment_hooks(prefix_path, pkg_name)
            assert isinstance(hooks, Iterable), \
                'create_environment_hooks() should return an iterable'
        except Exception as e:  # noqa: F841
            # catch exceptions raised in environment extension
            exc = traceback.format_exc()
            logger.error(
                'Exception in environment extension '
                "'{extension.ENVIRONMENT_NAME}': {e}\n{exc}"
                .format_map(locals()))
            # skip failing extension, continue with next one
            continue
        all_hooks += hooks
    return all_hooks



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import asyncio
from collections import OrderedDict
import os
import signal
import sys
from threading import Thread
import time

from colcon_core.executor import Job
from colcon_core.executor import OnError
from colcon_core.executor.sequential import SequentialExecutor
import pytest

ran_jobs = []


class Job1(Job):

    def __init__(self):
        super().__init__(
            identifier='job1', dependencies=set(), task=None,
            task_context=None)

    async def __call__(self, *args, **kwargs):
        global ran_jobs
        ran_jobs.append(self.identifier)


class Job2(Job):

    def __init__(self):
        super().__init__(
            identifier='job2', dependencies=set(), task=None,
            task_context=None)

    async def __call__(self, *args, **kwargs):
        return 2


class Job3(Job):

    def __init__(self):
        super().__init__(
            identifier='job3', dependencies=set(), task=None,
            task_context=None)

    async def __call__(self, *args, **kwargs):
        raise RuntimeError('custom exception')


class Job4(Job):

    def __init__(self):
        super().__init__(
            identifier='job4', dependencies=set(), task=None,
            task_context=None)

    async def __call__(self, *args, **kwargs):
        global ran_jobs
        ran_jobs.append(self.identifier)


class Job5(Job):

    def __init__(self):
        super().__init__(
            identifier='job5', dependencies=set(), task=None,
            task_context=None)

    async def __call__(self, *args, **kwargs):
        return 5


class Job6(Job):

    def __init__(self):
        super().__init__(
            identifier='job6', dependencies=('job2', ), task=None,
            task_context=None)

    async def __call__(self, *args, **kwargs):
        global ran_jobs
        ran_jobs.append(self.identifier)


class Job7(Job):

    def __init__(self):
        super().__init__(
            identifier='job7', dependencies=('job1', ), task=None,
            task_context=None)

    async def __call__(self, *args, **kwargs):
        global ran_jobs
        ran_jobs.append(self.identifier)


def test_sequential():
    global ran_jobs
    extension = SequentialExecutor()

    args = None
    jobs = OrderedDict()
    jobs['one'] = Job1()

    # success
    rc = extension.execute(args, jobs)
    assert rc == 0
    assert ran_jobs == ['job1']
    ran_jobs.clear()

    # return error code
    jobs['two'] = Job2()
    jobs['four'] = Job4()
    rc = extension.execute(args, jobs)
    assert rc == 2
    assert ran_jobs == ['job1']
    ran_jobs.clear()

    rc = extension.execute(args, jobs, on_error=OnError.skip_pending)
    assert rc == 2
    assert ran_jobs == ['job1']
    ran_jobs.clear()

    # continue after error, keeping first error code
    jobs['five'] = Job5()
    rc = extension.execute(args, jobs, on_error=OnError.continue_)
    assert rc == 2
    assert ran_jobs == ['job1', 'job4']
    ran_jobs.clear()

    # continue but skip downstream
    jobs['six'] = Job6()
    jobs['seven'] = Job7()
    rc = extension.execute(args, jobs, on_error=OnError.skip_downstream)
    assert rc == 2
    assert ran_jobs == ['job1', 'job4', 'job7']
    ran_jobs.clear()

    # exception
    jobs['two'] = Job3()
    rc = extension.execute(args, jobs)
    assert rc == 1
    assert ran_jobs == ['job1']
    ran_jobs.clear()


async def job8():
    global ran_jobs
    await asyncio.sleep(1)
    ran_jobs.append('job8')


def test_sequential_keyboard_interrupt():
    global ran_jobs

    if 'APPVEYOR' in os.environ:
        pytest.skip(
            'Skipping keyboard interrupt test since otherwise the prompt '
            "'Terminate batch job' blocks the build on AppVeyor")

    extension = SequentialExecutor()

    args = None
    jobs = OrderedDict()
    jobs['one'] = Job1()
    jobs['aborted'] = job8
    jobs['four'] = Job4()

    def delayed_sigint():
        time.sleep(0.1)
        os.kill(
            os.getpid(),
            signal.SIGINT if sys.platform != 'win32' else signal.CTRL_C_EVENT)
        if sys.platform == 'win32':
            os.kill(os.getpid(), signal.CTRL_C_EVENT)

    thread = Thread(target=delayed_sigint)
    try:
        thread.start()
        rc = extension.execute(args, jobs)
        assert rc == signal.SIGINT
    finally:
        thread.join()



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import os


def get_setup_data(pkg, env):
    """
    Get the options from the Python manifest.

    :param pkg: The package descriptor
    :param env: The environment
    :returns: The options
    :rtype: dict
    """
    # get data from pkg metadata if available
    key = 'get_python_setup_options'
    if key not in pkg.metadata:
        return {}
    return dict(pkg.metadata[key](env))


def get_data_files_mapping(data_files):
    """
    Transform the data_files structure into a dictionary.

    :param data_files: either a list of source files or
      a list of tuples where the first element is the destination path and
      the second element is a list of source files
    :returns: a dictionary mapping the source file to a destination file
    :rtype: dict
    """
    mapping = {}
    for data_file in data_files:
        if isinstance(data_file, tuple):
            assert len(data_file) == 2
            dest = data_file[0]
            assert not os.path.isabs(dest)
            sources = data_file[1]
            assert isinstance(sources, list)
            for source in sources:
                assert not os.path.isabs(source), \
                    "'data_files' must be relative, '{source}' is absolute" \
                    .format_map(locals())
                mapping[source] = os.path.join(dest, os.path.basename(source))
        else:
            assert not os.path.isabs(data_file)
            mapping[data_file] = os.path.basename(data_file)
    return mapping
